"""External Processor implementation for Envoy."""

import json

from grpc import ServicerContext

from langgate.core.logging import get_logger
from langgate.processor.envoy_extproc import (
    BaseExtProcService,
    CommonResponse,
    HttpBody,
    HttpHeaders,
)
from langgate.transform import LocalTransformerClient

logger = get_logger(__name__)

PROXY_HEADER = "x-langgate"
TARGET_ENDPOINT = "/v1"
ROUTE_HEADER = "x-route-to"


class LangGateExtProc(BaseExtProcService):
    """
    External Processor for Envoy.
    Implements the Envoy ext_proc External Processing Filter proto.

    This class processes HTTP requests and responses by implementing
    handlers for different parts of the request/response cycle.

    Note: This is a placeholder implementation that mimics the structure
    of a real ext_proc implementation but without direct imports from
    the envoy_extproc_sdk which needs to be updated.

    The envoy_extproc_sdk is based on the gRPC generated code from
    Envoy's External Processing filter proto definitions, defined in:
    https://github.com/envoyproxy/envoy/blob/main/api/envoy/service/ext_proc/v3/external_processor.proto

    TODO: Upgrade Envoy in https://github.com/Tanantor/envoy-extproc-sdk
    to address CVEs before implementing this service.
    This will involve significant refactoring of envoy-extproc-sdk to
    implement an architecture like the awesome Go ext_proc SDK by the same author.

    We may opt to use Go for this implementation instead of Python.
    Alternatively we may even use PyO3 for Rust-Python bindings.

    Tools like https://github.com/go-python/gopy for generating Python extension modules from Go
    could be used to create a Python wrapper around the Go tranformers for the local transformer setup.
    Another approach is building Go code as shared libraries using cgo and loading them in Python via ctypes.
    This would require packaging pre-compiled binaries for different platforms to maintain the simplicity
    of pip installation for end users.
    """

    def __init__(self, name: str = "LangGateExtProc"):
        """Initialize the processor."""
        super().__init__(name=name)
        self.transformer = LocalTransformerClient()
        logger.info("langgate_ext_proc_service_initialized")

    async def process_request_headers(
        self,
        headers: HttpHeaders,
        context: ServicerContext,
        request: dict,
        response: CommonResponse,
    ) -> CommonResponse:
        # Skip processing if not specifically targeting v1 endpoint
        if not request.get("path", "").startswith(TARGET_ENDPOINT):
            return response

        # Store content-type to check if JSON in body phase
        content_type = self.get_header(headers, "content-type")
        request["content_type"] = content_type

        # Store the original path for possible rewriting later
        request["original_path"] = request.get("path", "")
        self.logger.debug(f"Setting original path context: {request['original_path']}")

        # Replace the Authorization header
        # This is a placeholder for the actual implementation
        # TODO: We need to implement how we are actually doing this with the transformer
        auth_header = self.get_header(headers, "authorization")
        if auth_header:
            self.remove_header(response, "authorization")
            self.add_header(response, "authorization", "Bearer sk-proxy-key-replaced")
            self.logger.debug("Replaced Authorization header")

        return response

    async def process_request_body(
        self,
        body: HttpBody,
        context: ServicerContext,
        request: dict,
        response: CommonResponse,
    ) -> CommonResponse:
        original_path = request.get("original_path", "")
        if not original_path.startswith(TARGET_ENDPOINT):
            return response

        if (
            request.get("content_type")
            and "application/json" in request["content_type"]
        ):
            try:
                body_str = body.body.decode("utf-8")
                self.logger.debug(f"ORIGINAL BODY: {body_str}")
                json_body = json.loads(body_str)

                # Track modifications
                modified = False

                # Extract and process the 'model' parameter if present
                if "model" in json_body:
                    original_model = json_body["model"]
                    request["original_model"] = original_model

                    # Transform parameters using our transformer
                    _, transformed_params = await self.transformer.get_params(
                        original_model, json_body
                    )

                    # Replace the original JSON with transformed parameters
                    modified = True
                    self.logger.debug(
                        f"Transforming parameters for model: {original_model}"
                    )

                    target_route = transformed_params.get("base_url", "")
                    self.add_header(response, ROUTE_HEADER, target_route)

                    # We can't directly set host/authority headers,
                    # but Envoy will use x-route-to for host rewriting
                    # based on our route configuration's host_rewrite_header

                    # Save target route for path rewriting
                    request["target_route"] = target_route

                    # Clear the route cache for the current client request.
                    # We modified headers that are used to calculate the route.
                    response.clear_route_cache = True  # type: ignore

                if modified:
                    # TODO: transformed params will contain some headers, not all of it is to be sent as body
                    # the current transformer is returning params for a LangChain BaseChatModel.
                    new_body = json.dumps(transformed_params).encode("utf-8")

                    self.logger.debug(f"MODIFIED BODY SENT: {new_body.decode('utf-8')}")

                    response.body_mutation.body = new_body  # type: ignore

                    # When working with Envoy ExtProc, we don't set the content-length header
                    # as Envoy uses transfer-encoding: chunked when body is modified.
                    self.add_header(response, "x-content-length", str(len(new_body)))
                    request["body_modified"] = True

            except (json.JSONDecodeError, UnicodeDecodeError):
                self.logger.warning("Failed to decode JSON body")
        return response

    async def process_response_headers(
        self,
        headers: HttpHeaders,
        context: ServicerContext,
        request: dict,
        response: CommonResponse,
    ) -> CommonResponse:
        original_path = request.get("original_path", "")
        if not original_path.startswith(TARGET_ENDPOINT):
            return response

        self.logger.debug(
            f"""Processing response headers for path: {original_path}
            with context: {request} and headers: {headers}"""
        )
        # Add a marker header to indicate this request was proxied
        self.add_header(response, PROXY_HEADER, "true")

        return response



"""Tests for environment configuration."""

import os
from pathlib import Path
from unittest import mock

import pytest
import yaml
from pydantic import ValidationError

from langgate.registry.config import RegistryConfig
from langgate.registry.models import ModelRegistry
from tests.utils.config_utils import config_path_resolver
from tests.utils.registry_utils import (
    patch_model_registry,
    prevent_registry_config_loading,
)


@pytest.mark.parametrize(
    "source,expected_path",
    [
        ("arg", "/arg/path/models.json"),
        ("env", "/env/path/models.json"),
        ("cwd", "langgate_models.json"),
        ("package_dir", "default_models.json"),
    ],
    ids=["arg_path", "env_var", "cwd_path", "package_dir_path"],
)
def test_registry_config_models_json_paths(source, expected_path):
    """Test path resolution for models JSON file with different sources."""
    # Reset singleton for each case
    ModelRegistry._instance = None

    # Use our unified config path resolver
    with (
        prevent_registry_config_loading(),
        config_path_resolver(source, "models_json", expected_path),
    ):
        # For arg source, we need to explicitly pass the path
        if source == "arg":
            config = RegistryConfig(models_data_path=Path(expected_path))
        else:
            config = RegistryConfig()

        assert expected_path in str(config.models_data_path)


@pytest.mark.parametrize(
    "source,expected_path",
    [
        ("arg", "/arg/path/config.yaml"),
        ("env", "/env/path/config.yaml"),
        ("cwd", "langgate_config.yaml"),
        ("package_dir", "default_config.yaml"),
    ],
    ids=["arg_path", "env_var", "cwd_path", "package_dir_path"],
)
def test_registry_config_yaml_config_paths(source, expected_path):
    """Test path resolution for config YAML file with different sources."""
    # Reset singleton for each case
    ModelRegistry._instance = None

    with (
        prevent_registry_config_loading(),
        config_path_resolver(source, "config_yaml", expected_path),
    ):
        # For arg source, we need to explicitly pass the path
        if source == "arg":
            config = RegistryConfig(config_path=Path(expected_path))
        else:
            config = RegistryConfig()

        assert expected_path in str(config.config_path)


@pytest.mark.parametrize(
    "source,expected_path",
    [
        ("arg", "/arg/path/.env"),
        ("env", "/env/path/.env"),
        ("cwd", ".env"),
    ],
    ids=["arg_path", "env_var", "cwd_path"],
)
def test_registry_config_env_file_paths(source, expected_path):
    """Test path resolution for .env file with different sources."""
    # Reset singleton for each case
    ModelRegistry._instance = None

    with (
        prevent_registry_config_loading(),
        config_path_resolver(source, "env_file", expected_path),
    ):
        # For arg source, we need to explicitly pass the path
        if source == "arg":
            config = RegistryConfig(env_file_path=Path(expected_path))
        else:
            config = RegistryConfig()

        assert expected_path in str(config.env_file_path)


def test_registry_config_env_path_vars():
    """Test that environment variables set the correct paths."""
    env_vars = {
        "LANGGATE_MODELS": "/custom/path/langgate_models.json",
        "LANGGATE_CONFIG": "/custom/path/langgate_config.yaml",
        "LANGGATE_ENV_FILE": "/custom/path/.env",
    }
    with patch_model_registry(env_vars):
        registry = ModelRegistry()
        assert (
            str(registry.config.models_data_path) == "/custom/path/langgate_models.json"
        )
        assert str(registry.config.config_path) == "/custom/path/langgate_config.yaml"
        assert str(registry.config.env_file_path) == "/custom/path/.env"


def test_registry_config_without_env_file(mock_registry_files: dict[str, Path]):
    """Test that ModelRegistry works when .env file doesn't exist."""
    # Use the proper fixture but with non-existent .env file path
    non_existent_env = mock_registry_files["env_file"].parent / "nonexistent.env"

    with mock.patch.dict(
        os.environ,
        {
            "LANGGATE_CONFIG": str(mock_registry_files["config_yaml"]),
            "LANGGATE_MODELS": str(mock_registry_files["models_json"]),
            "LANGGATE_ENV_FILE": str(non_existent_env),
        },
    ):
        # Reset the singleton for environment variables to take effect
        ModelRegistry._instance = None

        # This should work without a .env file
        registry = ModelRegistry()

        # Verify models loaded correctly
        models = registry.list_llms()
        assert len(models) > 0
        assert any(model.id for model in models)


# Models merge mode tests


def test_merge_mode_default_behavior(
    mock_default_models_file: Path,
    mock_user_models_file: Path,
    mock_config_file: Path,
):
    """Test that merge mode is the default and merges models correctly."""
    with mock.patch("langgate.registry.config.importlib.resources") as mock_resources:
        # Mock the registry resources to point to our default models file
        mock_files = mock.MagicMock()
        mock_files.joinpath.return_value = mock_default_models_file
        mock_resources.files.return_value = mock_files

        config = RegistryConfig(
            models_data_path=mock_user_models_file, config_path=mock_config_file
        )

        # Should use merge mode by default
        assert config.models_merge_mode == "merge"

        # Should contain both default and user models
        assert len(config.models_data) == 3  # 2 default + 1 new custom

        # User model should override default
        assert config.models_data["openai/gpt-4o"]["name"] == "GPT-4o Custom"

        # Default model should remain
        assert (
            config.models_data["anthropic/claude-3-sonnet"]["name"]
            == "Claude 3 Sonnet Default"
        )

        # Custom model should be added
        assert config.models_data["custom/my-model"]["name"] == "My Custom Model"


def test_explicit_merge_mode(
    mock_default_models_file: Path,
    mock_user_models_file: Path,
    tmp_path: Path,
    merge_config_data: dict,
):
    """Test explicit merge mode configuration."""
    # Create config file with explicit merge mode
    config_file = tmp_path / "config.yaml"
    with open(config_file, "w") as f:
        yaml.dump(merge_config_data, f)

    with mock.patch("langgate.registry.config.importlib.resources") as mock_resources:
        # Mock the registry resources to point to our default models file
        mock_files = mock.MagicMock()
        mock_files.joinpath.return_value = mock_default_models_file
        mock_resources.files.return_value = mock_files

        config = RegistryConfig(
            models_data_path=mock_user_models_file, config_path=config_file
        )

        assert config.models_merge_mode == "merge"
        assert len(config.models_data) == 3
        assert config.models_data["openai/gpt-4o"]["name"] == "GPT-4o Custom"


def test_replace_mode(
    mock_default_models_file: Path,
    mock_user_models_file: Path,
    tmp_path: Path,
    replace_config_data: dict,
):
    """Test replace mode (legacy behavior)."""
    # Create config file with replace mode
    config_file = tmp_path / "config.yaml"
    with open(config_file, "w") as f:
        yaml.dump(replace_config_data, f)

    with mock.patch("langgate.registry.config.importlib.resources") as mock_resources:
        # Mock the registry resources to point to our default models file
        mock_files = mock.MagicMock()
        mock_files.joinpath.return_value = mock_default_models_file
        mock_resources.files.return_value = mock_files

        config = RegistryConfig(
            models_data_path=mock_user_models_file, config_path=config_file
        )

        assert config.models_merge_mode == "replace"
        # Should only contain user models
        assert len(config.models_data) == 2
        assert "anthropic/claude-3-sonnet" not in config.models_data
        assert config.models_data["openai/gpt-4o"]["name"] == "GPT-4o Custom"
        assert config.models_data["custom/my-model"]["name"] == "My Custom Model"


def test_extend_mode_success(
    mock_default_models_file: Path,
    tmp_path: Path,
    extend_config_data: dict,
    user_models_no_conflicts: dict,
):
    """Test extend mode with no conflicts."""
    # Create user models file with no conflicts
    user_models_file = tmp_path / "user_models.json"
    with open(user_models_file, "w") as f:
        import json

        json.dump(user_models_no_conflicts, f)

    # Create config file with extend mode
    config_file = tmp_path / "config.yaml"
    with open(config_file, "w") as f:
        yaml.dump(extend_config_data, f)

    with mock.patch("langgate.registry.config.importlib.resources") as mock_resources:
        # Mock the registry resources to point to our default models file
        mock_files = mock.MagicMock()
        mock_files.joinpath.return_value = mock_default_models_file
        mock_resources.files.return_value = mock_files

        config = RegistryConfig(
            models_data_path=user_models_file, config_path=config_file
        )

        assert config.models_merge_mode == "extend"
        # Should contain all models (2 default + 1 custom)
        assert len(config.models_data) == 3
        assert config.models_data["openai/gpt-4o"]["name"] == "GPT-4o Default"
        assert (
            config.models_data["anthropic/claude-3-sonnet"]["name"]
            == "Claude 3 Sonnet Default"
        )
        assert config.models_data["custom/my-model"]["name"] == "My Custom Model"


def test_extend_mode_conflict_error(
    mock_default_models_file: Path,
    mock_user_models_file: Path,
    tmp_path: Path,
    extend_config_data: dict,
):
    """Test extend mode raises error on conflicts."""
    # Create config file with extend mode
    config_file = tmp_path / "config.yaml"
    with open(config_file, "w") as f:
        yaml.dump(extend_config_data, f)

    with mock.patch("langgate.registry.config.importlib.resources") as mock_resources:
        # Mock the registry resources to point to our default models file
        mock_files = mock.MagicMock()
        mock_files.joinpath.return_value = mock_default_models_file
        mock_resources.files.return_value = mock_files

        with pytest.raises(ValueError) as exc_info:
            RegistryConfig(
                models_data_path=mock_user_models_file, config_path=config_file
            )

        assert "Model ID conflicts found in extend mode" in str(exc_info.value)
        assert "openai/gpt-4o" in str(exc_info.value)


def test_no_user_models_file(
    mock_default_models_file: Path,
    mock_config_file: Path,
):
    """Test behavior when no user models file exists."""
    non_existent_user_file = Path("/non/existent/path.json")

    with mock.patch("langgate.registry.config.importlib.resources") as mock_resources:
        # Mock the registry resources to point to our default models file
        mock_files = mock.MagicMock()
        mock_files.joinpath.return_value = mock_default_models_file
        mock_resources.files.return_value = mock_files

        config = RegistryConfig(
            models_data_path=non_existent_user_file, config_path=mock_config_file
        )

        # Should only contain default models
        assert len(config.models_data) == 2
        assert config.models_data["openai/gpt-4o"]["name"] == "GPT-4o Default"
        assert (
            config.models_data["anthropic/claude-3-sonnet"]["name"]
            == "Claude 3 Sonnet Default"
        )


def test_invalid_merge_mode(
    mock_default_models_file: Path,
    mock_user_models_file: Path,
    tmp_path: Path,
    invalid_config_data: dict,
):
    """Test error handling for invalid merge mode."""
    # Create config file with invalid merge mode
    config_file = tmp_path / "config.yaml"
    with open(config_file, "w") as f:
        yaml.dump(invalid_config_data, f)

    with mock.patch("langgate.registry.config.importlib.resources") as mock_resources:
        # Mock the registry resources to point to our default models file
        mock_files = mock.MagicMock()
        mock_files.joinpath.return_value = mock_default_models_file
        mock_resources.files.return_value = mock_files

        # Should raise ValidationError from Pydantic during config validation
        with pytest.raises(ValidationError) as exc_info:
            RegistryConfig(
                models_data_path=mock_user_models_file, config_path=config_file
            )

        # Check the validation error message
        assert "Input should be 'merge', 'replace' or 'extend'" in str(exc_info.value)


def test_merge_mode_logging(
    mock_default_models_file: Path,
    mock_user_models_file: Path,
    mock_config_file: Path,
    caplog,
):
    """Test that merge operations are properly logged."""
    with mock.patch("langgate.registry.config.importlib.resources") as mock_resources:
        # Mock the registry resources to point to our default models file
        mock_files = mock.MagicMock()
        mock_files.joinpath.return_value = mock_default_models_file
        mock_resources.files.return_value = mock_files

        with caplog.at_level("DEBUG"):
            RegistryConfig(
                models_data_path=mock_user_models_file, config_path=mock_config_file
            )

        # Check that debug logging occurred for model override
        debug_messages = [
            record.message for record in caplog.records if record.levelname == "DEBUG"
        ]
        override_logs = [
            msg for msg in debug_messages if "overriding_default_model" in msg
        ]
        assert len(override_logs) > 0

        # Check that info logging occurred with merge mode
        info_messages = [
            record.message for record in caplog.records if record.levelname == "INFO"
        ]
        model_data_logs = [msg for msg in info_messages if "loaded_model_data" in msg]
        assert len(model_data_logs) > 0



"""Registry API fixtures."""

import json
import os
from collections.abc import Generator
from pathlib import Path
from unittest import mock

import pytest
import yaml

from langgate.core.models import LLMInfo
from langgate.registry.local import LocalRegistryClient
from langgate.registry.models import ModelRegistry
from tests.factories.models import LLMInfoFactory
from tests.mocks.registry_mocks import CustomLocalRegistryClient


@pytest.fixture
def mock_models_json(tmp_path: Path) -> Generator[Path]:
    """Create a mock langgate_models.json file for testing."""
    models_data = {
        "openai/gpt-4o": {
            "name": "GPT-4o",
            "mode": "chat",
            "service_provider": "openai",
            "model_provider": "openai",
            "model_provider_name": "OpenAI",
            "capabilities": {
                "supports_tools": True,
                "supports_parallel_tool_calls": True,
                "supports_vision": True,
                "supports_prompt_caching": True,
                "supports_response_schema": True,
                "supports_system_messages": True,
                "supports_tool_choice": True,
            },
            "context": {"max_input_tokens": 128000, "max_output_tokens": 16384},
            "costs": {
                "input_cost_per_token": "0.0000025",
                "output_cost_per_token": "0.00001",
                "input_cost_per_token_batches": "0.00000125",
                "output_cost_per_token_batches": "0.000005",
                "cache_read_input_token_cost": "0.00000125",
                "input_cost_per_image": "0.003613",
            },
            "description": "OpenAI's GP model",
            "_last_updated": "2025-03-21T21:40:54.742453+00:00",
            "_data_source": "openrouter",
            "_last_updated_from_id": "openai/gpt-4o",
        },
        "anthropic/claude-sonnet-4-0": {
            "name": "Claude-4 Sonnet",
            "mode": "chat",
            "service_provider": "anthropic",
            "model_provider": "anthropic",
            "model_provider_name": "Anthropic",
            "capabilities": {
                "supports_tools": True,
                "supports_vision": True,
                "supports_prompt_caching": True,
                "supports_response_schema": True,
                "supports_assistant_prefill": True,
                "supports_tool_choice": True,
            },
            "context": {"max_input_tokens": 200000, "max_output_tokens": 128000},
            "costs": {
                "input_cost_per_token": "0.000003",
                "output_cost_per_token": "0.000015",
                "cache_read_input_token_cost": "3E-7",
                "cache_creation_input_token_cost": "0.00000375",
                "input_cost_per_image": "0.0048",
            },
            "description": "Anthropic's Claude 4 Sonnet model",
            "openrouter_model_id": "anthropic/claude-4-sonnet",
            "_last_updated": "2025-03-21T21:40:54.743326+00:00",
            "_data_source": "openrouter",
            "_last_updated_from_id": "anthropic/claude-4-sonnet",
        },
        "openai/dall-e-3": {
            "name": "DALL-E 3",
            "mode": "image",
            "service_provider": "openai",
            "model_provider": "openai",
            "model_provider_name": "OpenAI",
            "description": "OpenAI's DALL-E 3 model.",
            "costs": {
                "image_generation": {
                    "quality_tiers": {
                        "standard": {
                            "1024x1024": 0.04,
                            "1024x1792": 0.08,
                            "1792x1024": 0.08,
                        },
                        "hd": {"1024x1024": 0.08, "1024x1792": 0.12, "1792x1024": 0.12},
                    }
                }
            },
        },
        "replicate/black-forest-labs/flux-dev": {
            "name": "FLUX.1 [dev]",
            "mode": "image",
            "service_provider": "replicate",
            "model_provider": "black-forest-labs",
            "model_provider_name": "Black Forest Labs",
            "description": "The FLUX.1 dev model from Black Forest Labs.",
            "costs": {"image_generation": {"flat_rate": 0.025}},
        },
        "replicate/stability-ai/stable-diffusion-3.5-large": {
            "name": "SD 3.5 Large",
            "mode": "image",
            "service_provider": "replicate",
            "model_provider": "stability-ai",
            "model_provider_name": "Stability AI",
            "description": "Stability AI's Stable Diffusion 3.5 Large model.",
            "costs": {"image_generation": {"flat_rate": 0.065}},
        },
    }
    models_json_path = tmp_path / "langgate_models.json"
    with open(models_json_path, "w") as f:
        json.dump(models_data, f)

    yield models_json_path


@pytest.fixture
def mock_config_yaml(tmp_path: Path) -> Generator[Path]:
    """Create a mock langgate_config.yaml file for testing."""
    config_data = {
        "app_config": {
            "PROJECT_NAME": "LangGate Test",
            "CORS_ORIGINS": ["http://localhost"],
            "HTTPS": False,
        },
        "default_params": {
            "text": {
                "temperature": 0.7,
            },
        },
        "services": {
            "openai": {
                "api_key": "${OPENAI_API_KEY}",
                "base_url": "https://api.openai.com/v1",
                "default_params": {
                    "text": {
                        "max_tokens": 1000,
                    },
                },
            },
            "anthropic": {
                "api_key": "${ANTHROPIC_API_KEY}",
                "base_url": "https://api.anthropic.com",
                "default_params": {
                    "text": {
                        "max_tokens": 2000,
                    },
                },
                "model_patterns": {
                    "reasoning": {
                        "default_params": {"max_tokens": 64000},
                        "override_params": {
                            "thinking": {
                                "type": "enabled",
                            },
                        },
                        "remove_params": ["temperature"],
                        "rename_params": {"reasoning": "thinking"},
                    }
                },
            },
            "openrouter": {
                "api_key": "${OPENROUTER_API_KEY}",
                "base_url": "https://api.openrouter.ai/v1",
                "api_format": "openai",
                "default_params": {
                    "text": {
                        "tiktoken_model_name": "gpt-4o",
                    },
                },
            },
            "xai": {
                "api_key": "${XAI_API_KEY}",
                "base_url": "https://api.x.ai/v1",
                "api_format": "openai",
            },
            "fireworks_ai": {
                "api_key": "${FIREWORKS_API_KEY}",
                "base_url": "https://api.fireworks.ai/inference/v1",
                "api_format": "openai",
                "default_params": {
                    "text": {
                        "tiktoken_model_name": "gpt-4o",
                    },
                },
            },
            "gemini": {
                "api_key": "${GEMINI_API_KEY}",
            },
            "mistralai": {
                "api_key": "${MISTRAL_API_KEY}",
                "base_url": "https://api.mistral.ai/v1",
            },
            "replicate": {
                "api_key": "${REPLICATE_API_KEY}",
            },
        },
        "models": {
            "text": [
                {
                    "id": "gpt-4o",
                    "name": "GPT-4o",
                    "service": {
                        "provider": "openai",
                        "model_id": "gpt-4o",
                    },
                },
                {
                    "id": "anthropic/claude-sonnet-4",
                    "name": "Claude-3.7 Sonnet",
                    "service": {
                        "provider": "anthropic",
                        "model_id": "claude-sonnet-4",
                    },
                    "remove_params": ["response_format", "reasoning"],
                    "rename_params": {"stop": "stop_sequences"},
                },
                {
                    "id": "anthropic/claude-sonnet-4-reasoning",
                    "name": "Claude-3.7 Sonnet R",
                    "description": "Claude 3.7 Sonnet with reasoning",
                    "service": {
                        "provider": "anthropic",
                        "model_id": "claude-sonnet-4",
                    },
                    "remove_params": ["response_format"],
                    "rename_params": {"stop": "stop_sequences"},
                    "override_params": {
                        "thinking": {
                            "budget_tokens": 1024,
                        }
                    },
                },
                {
                    "id": "google/gemma-3-27b-it",
                    "service": {
                        "provider": "openrouter",
                        "model_id": "google/gemma-3-27b-it:free",
                    },
                },
                {
                    "id": "xai/grok-3",
                    "service": {
                        "provider": "xai",
                        "model_id": "grok-3-latest",
                    },
                },
                {
                    "id": "deepseek/deepseek-r1",
                    "service": {
                        "provider": "fireworks_ai",
                        "model_id": "accounts/fireworks/models/deepseek-r1",
                    },
                },
                {
                    "id": "google/gemini-2.5-pro",
                    "service": {
                        "provider": "gemini",
                        "model_id": "gemini-2.5-pro-preview",
                    },
                },
                {
                    "id": "mistralai/magistral-medium-latest",
                    "service": {
                        "provider": "mistralai",
                        "model_id": "magistral-medium-latest",
                    },
                },
            ],
            "image": [
                {
                    "id": "openai/dall-e-3",
                    "name": "DALL-E 3",
                    "service": {
                        "provider": "openai",
                        "model_id": "dall-e-3",
                    },
                },
                {
                    "id": "black-forest-labs/flux-dev",
                    "name": "FLUX.1 [dev]",
                    "service": {
                        "provider": "replicate",
                        "model_id": "black-forest-labs/flux-dev",
                    },
                    "default_params": {
                        "disable_safety_checker": True,
                    },
                },
                {
                    "id": "stability-ai/sd-3.5-large",
                    "name": "Stable Diffusion 3.5 Large",
                    "service": {
                        "provider": "replicate",
                        "model_id": "stability-ai/stable-diffusion-3.5-large",
                    },
                },
            ],
        },
    }
    config_yaml_path = tmp_path / "langgate_config.yaml"
    with open(config_yaml_path, "w") as f:
        yaml.dump(config_data, f)

    yield config_yaml_path


@pytest.fixture
def mock_env_file(tmp_path: Path) -> Generator[Path]:
    """Create a mock .env file for testing."""
    env_path = tmp_path / ".env"
    with open(env_path, "w") as f:
        f.write("OPENAI_API_KEY=sk-test-123\n")
        f.write("ANTHROPIC_API_KEY=sk-ant-test-123\n")
        f.write("OPENROUTER_API_KEY=sk-or-test-123\n")
        f.write("XAI_API_KEY=xai-test-123\n")
        f.write("FIREWORKS_API_KEY=fw-test-123\n")
        f.write("GEMINI_API_KEY=gm-test-123\n")
        f.write("MISTRAL_API_KEY=ms-test-123\n")
        f.write("REPLICATE_API_KEY=r8-test-123\n")
        f.write("SECRET_KEY=test-secret-key\n")

    yield env_path


@pytest.fixture
def mock_registry_files(
    mock_models_json: Path, mock_config_yaml: Path, mock_env_file: Path
) -> Generator[dict[str, Path]]:
    """Combine all mock registry files for testing."""
    result_dict = {
        "models_json": mock_models_json,
        "config_yaml": mock_config_yaml,
        "env_file": mock_env_file,
    }
    yield result_dict


@pytest.fixture
def mock_models_path_in_env(mock_models_json: Path) -> Generator[dict[str, str]]:
    """Mock only the langgate_models.json environment variable."""
    with mock.patch.dict(os.environ, {"LANGGATE_MODELS": str(mock_models_json)}):
        yield {"LANGGATE_MODELS": str(mock_models_json)}


@pytest.fixture
def mock_config_path_in_env(mock_config_yaml: Path) -> Generator[dict[str, str]]:
    """Mock only the langgate_config.yaml environment variable."""
    with mock.patch.dict(os.environ, {"LANGGATE_CONFIG": str(mock_config_yaml)}):
        yield {"LANGGATE_CONFIG": str(mock_config_yaml)}


@pytest.fixture
def mock_env_file_path_in_env(mock_env_file: Path) -> Generator[dict[str, str]]:
    """Mock only the .env file environment variable."""
    with mock.patch.dict(os.environ, {"LANGGATE_ENV_FILE": str(mock_env_file)}):
        yield {"LANGGATE_ENV_FILE": str(mock_env_file)}


@pytest.fixture
def mock_env_vars_from_env_file(mock_env_file: Path) -> Generator[dict[str, str]]:
    """Mock environment variables loaded from the .env file."""
    env_vars = {}
    with open(mock_env_file) as f:
        for _line in f:
            line = _line.strip()
            if line and not line.startswith("#"):
                key, value = line.split("=", 1)
                env_vars[key] = value

    with mock.patch.dict(os.environ, env_vars):
        yield env_vars


@pytest.fixture
def mock_all_env_vars(
    mock_models_json: Path,
    mock_config_yaml: Path,
    mock_env_file: Path,
    mock_env_vars_from_env_file: dict[str, str],
) -> Generator[dict[str, str]]:
    """Mock all registry-related environment variables."""
    env_vars = mock_env_vars_from_env_file.copy()
    env_vars.update(
        {
            "LANGGATE_MODELS": str(mock_models_json),
            "LANGGATE_CONFIG": str(mock_config_yaml),
            "LANGGATE_ENV_FILE": str(mock_env_file),
        }
    )

    with mock.patch.dict(os.environ, env_vars):
        yield env_vars


@pytest.fixture
def model_registry(mock_all_env_vars: dict[str, str]) -> Generator[ModelRegistry]:
    """Create a ModelRegistry instance with mock files."""
    ModelRegistry._instance = None

    # Create a fresh registry instance
    registry = ModelRegistry()
    yield registry
    ModelRegistry._instance = None


@pytest.fixture
def mock_llm_info() -> LLMInfo:
    """Create a mock LLMInfo instance."""
    return LLMInfoFactory.create()


@pytest.fixture
def local_registry_client() -> Generator[LocalRegistryClient]:
    """Return a LocalRegistryClient instance."""
    original_instance = LocalRegistryClient._instance
    LocalRegistryClient._instance = None

    client = LocalRegistryClient()

    yield client

    LocalRegistryClient._instance = original_instance


@pytest.fixture
def custom_local_registry_client() -> Generator[CustomLocalRegistryClient]:
    """Return a CustomLocalRegistryClient instance."""
    client = CustomLocalRegistryClient()

    yield client


@pytest.fixture
def default_models_data() -> dict:
    """Sample default models data for merge testing."""
    return {
        "openai/gpt-4o": {
            "name": "GPT-4o Default",
            "service_provider": "openai",
            "model_provider": "openai",
            "description": "Default GPT-4o model",
        },
        "anthropic/claude-3-sonnet": {
            "name": "Claude 3 Sonnet Default",
            "service_provider": "anthropic",
            "model_provider": "anthropic",
            "description": "Default Claude 3 Sonnet model",
        },
    }


@pytest.fixture
def user_models_data() -> dict:
    """Sample user models data for merge testing."""
    return {
        "openai/gpt-4o": {
            "name": "GPT-4o Custom",
            "service_provider": "openai",
            "model_provider": "openai",
            "description": "Custom GPT-4o configuration",
        },
        "custom/my-model": {
            "name": "My Custom Model",
            "service_provider": "custom",
            "model_provider": "custom",
            "description": "My custom model",
        },
    }


@pytest.fixture
def base_config_data() -> dict:
    """Base configuration data for merge testing."""
    return {
        "default_params": {},
        "services": {
            "openai": {
                "api_key": "test-key",
                "base_url": "https://api.openai.com/v1",
            },
            "anthropic": {
                "api_key": "test-key",
                "base_url": "https://api.anthropic.com",
            },
            "custom": {"api_key": "test-key", "base_url": "https://api.custom.com"},
        },
        "models": {"text": []},
        "app_config": {},
    }


@pytest.fixture
def merge_config_data(base_config_data: dict) -> dict:
    """Configuration data with merge mode set."""
    config_data = base_config_data.copy()
    config_data["models_merge_mode"] = "merge"
    return config_data


@pytest.fixture
def replace_config_data(base_config_data: dict) -> dict:
    """Configuration data with replace mode set."""
    config_data = base_config_data.copy()
    config_data["models_merge_mode"] = "replace"
    return config_data


@pytest.fixture
def extend_config_data(base_config_data: dict) -> dict:
    """Configuration data with extend mode set."""
    config_data = base_config_data.copy()
    config_data["models_merge_mode"] = "extend"
    return config_data


@pytest.fixture
def invalid_config_data(base_config_data: dict) -> dict:
    """Configuration data with invalid merge mode."""
    config_data = base_config_data.copy()
    config_data["models_merge_mode"] = "invalid"
    return config_data


@pytest.fixture
def user_models_no_conflicts() -> dict:
    """User models data with no conflicts for extend mode testing."""
    return {
        "custom/my-model": {
            "name": "My Custom Model",
            "service_provider": "custom",
            "model_provider": "custom",
            "description": "My custom model",
        }
    }


@pytest.fixture
def mock_default_models_file(tmp_path: Path, default_models_data: dict) -> Path:
    """Create a temporary default models JSON file."""
    default_models_file = tmp_path / "default_models.json"
    with open(default_models_file, "w") as f:
        json.dump(default_models_data, f)
    return default_models_file


@pytest.fixture
def mock_user_models_file(tmp_path: Path, user_models_data: dict) -> Path:
    """Create a temporary user models JSON file."""
    user_models_file = tmp_path / "user_models.json"
    with open(user_models_file, "w") as f:
        json.dump(user_models_data, f)
    return user_models_file


@pytest.fixture
def mock_config_file(tmp_path: Path, base_config_data: dict) -> Path:
    """Create a temporary config YAML file."""
    config_file = tmp_path / "config.yaml"
    with open(config_file, "w") as f:
        yaml.dump(base_config_data, f)
    return config_file



"""Integration tests for the models endpoint."""

from pathlib import Path
from unittest import mock

import pytest
from httpx import AsyncClient

LLMS_URL = "/models/llms"
IMAGE_MODELS_URL = "/models/images"


@pytest.mark.asyncio
async def test_get_models(registry_api_client: AsyncClient) -> None:
    """Test retrieving all LLMs."""
    response = await registry_api_client.get(LLMS_URL)
    assert response.status_code == 200

    data = response.json()
    assert isinstance(data, list)
    assert len(data) > 0

    for llm in data:
        assert isinstance(llm, dict)
        assert "id" in llm
        assert "name" in llm
        assert "provider" in llm
        assert "costs" in llm
        assert "capabilities" in llm
        assert "context_window" in llm

        provider = llm["provider"]
        assert "id" in provider
        assert "name" in provider


@pytest.mark.asyncio
async def test_get_model_by_id(registry_api_client: AsyncClient) -> None:
    """Test retrieving a specific LLM by ID."""
    # First get all LLMs
    response = await registry_api_client.get(LLMS_URL)
    assert response.status_code == 200

    llms = response.json()
    assert len(llms) > 0
    test_llm = llms[0]

    # Test API response for specific LLM
    response = await registry_api_client.get(f"{LLMS_URL}/{test_llm['id']}")
    assert response.status_code == 200

    llm = response.json()
    assert llm["id"] == test_llm["id"]
    assert llm["name"] == test_llm["name"]
    assert llm["provider"]["id"] == test_llm["provider"]["id"]

    # Test with non-existent LLM
    response = await registry_api_client.get(f"{LLMS_URL}/non-existent-llm")
    assert response.status_code == 404


@pytest.mark.asyncio
async def test_model_schema(registry_api_client: AsyncClient) -> None:
    """Test that the LLM schema validation works correctly."""
    # Get LLMs to verify schema
    response = await registry_api_client.get(LLMS_URL)
    assert response.status_code == 200

    llms = response.json()
    assert len(llms) > 0

    # Attempt to validate LLM structure using Pydantic
    for llm_data in llms:
        # We can't directly validate here, but we should check key fields
        assert isinstance(llm_data["id"], str)
        assert isinstance(llm_data["name"], str)

        # Check costs structure
        costs = llm_data["costs"]
        assert any(k.startswith("input_") for k in costs)
        assert any(k.startswith("output_") for k in costs)

        # Check capabilities structure
        capabilities = llm_data["capabilities"]
        if capabilities:
            assert all(k.startswith("supports_") for k in capabilities)

        # Check context window
        context = llm_data["context_window"]
        assert "max_input_tokens" in context
        assert "max_output_tokens" in context


@pytest.mark.asyncio
async def test_models_api_works_without_env_file(
    registry_api_client: AsyncClient,
) -> None:
    """Test that the LLMs API endpoints work without env files."""
    # Make Path.exists return False for .env files
    with mock.patch.object(Path, "exists") as mock_exists:

        def side_effect(path):
            # Return False for .env files, True for everything else
            return ".env" not in str(path)

        mock_exists.side_effect = side_effect

        # Test the LLMs route still returns successfully
        response = await registry_api_client.get(LLMS_URL)
        assert response.status_code == 200

        # Test that we can get one of the LLMs by ID
        llms = response.json()
        assert llms, "No LLMs found in registry"
        test_llm = llms[0]
        llm_id = test_llm["id"]

        response = await registry_api_client.get(f"{LLMS_URL}/{llm_id}")
        assert response.status_code == 200
        assert response.json()["id"] == llm_id


# Image model tests


@pytest.mark.asyncio
async def test_get_image_models(registry_api_client: AsyncClient) -> None:
    """Test retrieving all image models."""
    response = await registry_api_client.get(IMAGE_MODELS_URL)
    assert response.status_code == 200

    data = response.json()
    assert isinstance(data, list)
    assert len(data) > 0

    for model in data:
        assert isinstance(model, dict)
        assert "id" in model
        assert "name" in model
        assert "provider" in model
        assert "costs" in model
        assert "updated_dt" in model

        provider = model["provider"]
        assert "id" in provider
        assert "name" in provider

        costs = model["costs"]
        assert "image_generation" in costs
        image_generation = costs["image_generation"]

        # Verify exactly one pricing model is set (as per validation)
        pricing_models = [
            image_generation.get("flat_rate"),
            image_generation.get("quality_tiers"),
            image_generation.get("cost_per_megapixel"),
            image_generation.get("cost_per_second"),
        ]
        assert sum(p is not None for p in pricing_models) == 1


@pytest.mark.asyncio
async def test_get_image_model_by_id(registry_api_client: AsyncClient) -> None:
    """Test retrieving a specific image model by ID."""
    # First get all image models
    response = await registry_api_client.get(IMAGE_MODELS_URL)
    assert response.status_code == 200

    models = response.json()
    assert len(models) > 0
    test_model = models[0]

    # Test API response for specific image model
    response = await registry_api_client.get(f"{IMAGE_MODELS_URL}/{test_model['id']}")
    assert response.status_code == 200

    model = response.json()
    assert model["id"] == test_model["id"]
    assert model["name"] == test_model["name"]
    assert model["provider"]["id"] == test_model["provider"]["id"]

    # Test with non-existent image model
    response = await registry_api_client.get(
        f"{IMAGE_MODELS_URL}/non-existent-image-model"
    )
    assert response.status_code == 404


@pytest.mark.asyncio
async def test_image_model_schema(registry_api_client: AsyncClient) -> None:
    """Test that the image model schema validation works correctly."""
    # Get image models to verify schema
    response = await registry_api_client.get(IMAGE_MODELS_URL)
    assert response.status_code == 200

    models = response.json()
    assert len(models) > 0

    # Attempt to validate image model structure
    for model_data in models:
        # Check required fields
        assert isinstance(model_data["id"], str)
        assert isinstance(model_data["name"], str)

        # Check costs structure
        costs = model_data["costs"]
        assert "image_generation" in costs

        image_generation = costs["image_generation"]
        # Verify exactly one pricing model (business rule validation)
        pricing_models = [
            image_generation.get("flat_rate"),
            image_generation.get("quality_tiers"),
            image_generation.get("cost_per_megapixel"),
            image_generation.get("cost_per_second"),
        ]
        non_null_models = [p for p in pricing_models if p is not None]
        assert len(non_null_models) == 1, (
            f"Expected exactly 1 pricing model, got {len(non_null_models)}"
        )

        # Check provider structure
        provider = model_data["provider"]
        assert "id" in provider
        assert "name" in provider


@pytest.mark.asyncio
async def test_image_models_api_works_without_env_file(
    registry_api_client: AsyncClient,
) -> None:
    """Test that the image models API endpoints work without env files."""
    # Make Path.exists return False for .env files
    with mock.patch.object(Path, "exists") as mock_exists:

        def side_effect(path):
            # Return False for .env files, True for everything else
            return ".env" not in str(path)

        mock_exists.side_effect = side_effect

        # Test the image models route still returns successfully
        response = await registry_api_client.get(IMAGE_MODELS_URL)
        assert response.status_code == 200

        # Test that we can get one of the image models by ID
        models = response.json()
        assert models, "No image models found in registry"
        test_model = models[0]
        model_id = test_model["id"]

        response = await registry_api_client.get(f"{IMAGE_MODELS_URL}/{model_id}")
        assert response.status_code == 200
        assert response.json()["id"] == model_id



"""Test utilities for configuration path resolution testing."""

import os
from contextlib import ExitStack, contextmanager
from pathlib import Path
from typing import Literal
from unittest import mock

from tests.mocks.registry_mocks import create_mock_config


@contextmanager
def patch_load_yaml_config(custom_config=None):
    """Patch the load_yaml_config function to return a mock config."""
    mock_config = custom_config if custom_config else create_mock_config()
    with (
        mock.patch(
            "langgate.transform.local.load_yaml_config",
            return_value=mock_config,
        ),
        mock.patch(
            "langgate.registry.config.load_yaml_config",
            return_value=mock_config,
        ),
    ):
        yield mock_config


@contextmanager
def patch_config_paths_arg_source(arg_path: Path):
    """Mock path resolution for constructor arguments source.

    This utility patches only what's needed to test constructor argument paths.

    Args:
        arg_path: The path provided as constructor argument.
    """
    with mock.patch("pathlib.Path.exists", return_value=True), patch_load_yaml_config():
        yield


@contextmanager
def patch_config_paths_env_source(env_var: str, env_value: str):
    """Mock path resolution for environment variable source.

    This utility patches only what's needed to test environment variable paths.

    Args:
        env_var: The environment variable name to set.
        env_value: The value to set the environment variable to.
    """
    env_vars = {env_var: env_value}
    with (
        mock.patch.dict(os.environ, env_vars, clear=True),
        mock.patch("pathlib.Path.exists", return_value=True),
        patch_load_yaml_config(),
    ):
        yield


@contextmanager
def patch_config_paths_cwd_source(pattern: str):
    """Mock path resolution for current working directory source.

    This utility patches only what's needed to test CWD paths.

    Args:
        pattern: The filename pattern to match in the mocked CWD.
    """
    # Mock CWD to a known path
    with mock.patch("pathlib.Path.cwd", return_value=Path("/fake/cwd")):
        # Create a path-specific exists check
        def patched_exists(path_obj):
            path_str = str(path_obj)
            return pattern in path_str and "/fake/cwd" in path_str

        with (
            mock.patch.object(Path, "exists", patched_exists),
            patch_load_yaml_config(),
        ):
            yield


@contextmanager
def patch_config_paths_default_source():
    """Mock path resolution for package directory fallback source.

    This utility patches only what's needed to test package directory fallback paths.
    """
    # Make all explicit paths not exist to force default package dir
    with (
        mock.patch("pathlib.Path.exists", return_value=False),
        patch_load_yaml_config(),
    ):
        yield


@contextmanager
def config_path_resolver(
    source_type: Literal["arg", "env", "cwd", "package_dir"],
    file_type: Literal["models_json", "config_yaml", "env_file"],
    path_value: str | None = None,
):
    """Combined context manager for all path resolution types.

    This reuses the existing context managers but provides a more consistent interface.

    Args:
        source_type: The type of source to patch (arg, env, cwd, package_dir)
        file_type: The type of file being patched (models_json, config_yaml, env_file)
        path_value: The path value to use (required for all except package_dir)
    """
    env_var_map = {
        "models_json": "LANGGATE_MODELS",
        "config_yaml": "LANGGATE_CONFIG",
        "env_file": "LANGGATE_ENV_FILE",
    }

    default_pattern_map = {
        "models_json": "langgate_models.json",
        "config_yaml": "langgate_config.yaml",
        "env_file": ".env",
    }

    with ExitStack() as stack:
        if source_type == "arg":
            # For constructor argument paths, we just need to make Path.exists return True
            if path_value is None:
                raise ValueError("path_value cannot be None for arg source_type")
            stack.enter_context(patch_config_paths_arg_source(Path(path_value)))

        elif source_type == "env":
            # For environment variable paths
            if path_value is None:
                raise ValueError("path_value cannot be None for env source_type")
            env_var = env_var_map[file_type]
            stack.enter_context(patch_config_paths_env_source(env_var, path_value))

        elif source_type == "cwd":
            # For current working directory paths
            pattern = default_pattern_map[file_type]
            stack.enter_context(patch_config_paths_cwd_source(pattern))

        elif source_type == "package_dir":
            # For package directory fallback paths
            stack.enter_context(patch_config_paths_default_source())

        yield


@contextmanager
def prevent_server_config_loading():
    """Prevent actual config file loading in ServerConfig.

    Use this to avoid actual file operations during server config tests.
    """
    with mock.patch(
        "langgate.server.core.config.FixedYamlConfigSettingsSource.__call__",
        return_value={},
    ):
        yield



from collections.abc import Generator
from typing import Any

import pytest
import pytest_asyncio
from httpx import AsyncClient as AsyncTestClient
from openai import AsyncOpenAI

from langgate.registry import ModelRegistry
from langgate.server.core.config import settings
from tests.utils.utils import random_lower_string

pytest_plugins = [
    "tests.fixtures.registry_fixtures",
    "tests.fixtures.transform_fixtures",
    "tests.fixtures.client_fixtures",
]


@pytest.fixture(autouse=True)
def reset_singletons():
    """Reset singleton instances before and after each test to prevent test interference.
    This ensures tests running in parallel don't affect each other through shared state.
    """
    ModelRegistry._instance = None
    yield
    ModelRegistry._instance = None


@pytest_asyncio.fixture
def openai_client(client: AsyncTestClient) -> Generator[AsyncOpenAI, Any]:
    openai_client = AsyncOpenAI(
        api_key="",
        base_url=f"{client.base_url}",
        http_client=client,
    )
    yield openai_client


@pytest.fixture(scope="module")
def vcr_config():
    def before_record_cb(response: dict[str, Any]):
        if response["headers"].get("openai-organization"):
            response["headers"]["openai-organization"] = ["DUMMY"]
        return response

    return {
        "filter_headers": [
            ("authorization", "DUMMY"),
            ("x-api-key", "DUMMY"),
            ("openai-organization", "DUMMY"),
        ],
        "filter_query_parameters": [
            ("api_key", "DUMMY"),
            ("openai_api_key", "DUMMY"),
            ("client_secret", "DUMMY"),
        ],
        "filter_post_data_parameters": [
            ("previewToken", "DUMMY"),
            ("client_secret", "DUMMY"),
            ("refresh_token", "DUMMY"),
        ],
        "ignore_localhost": False,
        "ignore_hosts": [
            settings.TEST_SERVER_HOST,
            # tiktoken: https://community.openai.com/t/newconnectionerror-keeps-coming-up-over-a-tiktoken-file/670052/1
            "openaipublic.blob.core.windows.net",
        ],
        "record_mode": "once",  # delete cassettes or change to "all" to record new cassettes
        "before_record_response": before_record_cb,
    }


@pytest.fixture
def random_str() -> str:
    return random_lower_string()



"""Unit tests for LocalRegistryClient."""

from langgate.core.models import ImageModelInfo
from langgate.registry.local import BaseLocalRegistryClient, LocalRegistryClient
from tests.mocks.registry_mocks import CustomLLMInfo, CustomLocalRegistryClient
from tests.utils.registry_utils import patch_model_registry


def test_local_registry_client_is_singleton():
    """Test that LocalRegistryClient implements the singleton pattern."""
    with patch_model_registry(reset_singleton=True):
        # Reset singleton for test
        LocalRegistryClient._instance = None

        # Create instances
        client1 = LocalRegistryClient()
        client2 = LocalRegistryClient()

        # Verify they are the same instance
        assert client1 is client2

        # Verify they share the same registry
        assert id(client1.registry) == id(client2.registry)


def test_custom_client_type_parameter_extraction():
    """Test that type parameters are correctly extracted from subclasses."""
    with patch_model_registry():
        # CustomLocalRegistryClient subclasses BaseLocalRegistryClient with CustomLLMInfo
        client = CustomLocalRegistryClient()

        # Verify the generic type parameter extraction worked
        assert client.llm_info_cls is CustomLLMInfo


def test_explicit_model_class_parameter():
    """Test passing explicit llm_info_cls to BaseLocalRegistryClient."""
    with patch_model_registry():
        # When using directly with type parameter, llm_info_cls must be provided explicitly

        client = BaseLocalRegistryClient[CustomLLMInfo, ImageModelInfo](
            llm_info_cls=CustomLLMInfo
        )

        # Verify the llm_info_cls is correctly set
        assert client.llm_info_cls is CustomLLMInfo



#!/usr/bin/env python3
"""
Envoy configuration template renderer.
Renders Jinja2 templates with environment variables and outputs the result to a file.
"""

import argparse
import json
import os
import sys
from typing import Any

DEBUG = os.environ.get("DEBUG") == "1"

try:
    import jinja2
except ImportError:
    sys.stderr.write(
        'ERROR: Jinja2 library not found. Please install it with "pip install jinja2"\n'
    )
    sys.exit(1)


class TemplateVariableProvider:
    """Base class for providing template variables from different sources."""

    def get_variables(self) -> dict[str, Any]:
        """Get variables from the source."""
        raise NotImplementedError("Subclasses must implement get_variables")


class EnvironmentVariableProvider(TemplateVariableProvider):
    """Provides template variables from environment variables."""

    def __init__(self) -> None:
        """Initialize the provider."""

    @staticmethod
    def clean_value(value: str) -> str:
        """Clean environment variable values by removing unnecessary quotes."""
        if isinstance(value, str) and (
            (value.startswith('"') and value.endswith('"'))
            or (value.startswith("'") and value.endswith("'"))
        ):
            return value[1:-1]
        return value

    def get_variables(self) -> dict[str, Any]:
        """Get all environment variables as template variables."""
        variables = {}
        for key, value in os.environ.items():
            # Convert to lowercase for template use
            template_key = key.lower()
            clean_value = self.clean_value(value)
            variables[template_key] = clean_value
        return variables


class FileVariableProvider(TemplateVariableProvider):
    """Provides template variables from a JSON file."""

    def __init__(self, file_path: str) -> None:
        """Initialize with the path to the JSON file."""
        self.file_path = file_path

    def get_variables(self) -> dict[str, Any]:
        """Load variables from the JSON file."""
        try:
            with open(self.file_path) as f:
                return json.load(f)
        except (OSError, json.JSONDecodeError) as e:
            sys.stderr.write(
                f"ERROR: Failed to load variables from {self.file_path}: {e}\n"
            )
            sys.exit(1)


class TemplateRenderer:
    """Renders templates using Jinja2."""

    def __init__(self, template_file: str, output_file: str) -> None:
        """Initialize with template and output file paths."""
        self.template_file = template_file
        self.output_file = output_file
        self.variables: dict[str, Any] = {}

    def add_variables(self, variables: dict[str, Any]) -> None:
        """Add variables to be used in template rendering."""
        self.variables.update(variables)

    def set_default(self, key: str, value: Any) -> None:
        """Set a default value for a variable if it doesn't exist."""
        if key not in self.variables:
            self.variables[key] = value

    def render(self) -> None:
        """Render the template with the provided variables."""
        try:
            # Load the template
            with open(self.template_file) as f:
                template_content = f.read()

            template = jinja2.Template(template_content)
            rendered_content = template.render(**self.variables)

            with open(self.output_file, "w") as f:
                f.write(rendered_content)

            # Save a copy with the debug suffix if in debug mode
            if DEBUG:
                debug_output_file = f"{self.output_file}.debug"
                with open(debug_output_file, "w") as f:
                    f.write(rendered_content)
                print(f"Debug copy saved to {debug_output_file}")

            print(f"Successfully rendered {self.template_file} to {self.output_file}")

        except Exception as e:
            sys.stderr.write(f"ERROR: Failed to render template: {e}\n")
            sys.exit(1)


def main() -> None:
    """Main entry point for the script."""
    parser = argparse.ArgumentParser(
        description="Render Envoy configuration templates using environment variables."
    )
    parser.add_argument(
        "--template", "-t", required=True, help="Path to the template file"
    )
    parser.add_argument("--output", "-o", required=True, help="Path to the output file")
    parser.add_argument(
        "--vars", "-v", help="Path to JSON file with additional template variables"
    )

    args = parser.parse_args()

    renderer = TemplateRenderer(args.template, args.output)

    env_provider = EnvironmentVariableProvider()
    renderer.add_variables(env_provider.get_variables())

    # Add variables from file if provided (overrides environment vars)
    if args.vars:
        file_provider = FileVariableProvider(args.vars)
        renderer.add_variables(file_provider.get_variables())

    # Set default values for required variables
    renderer.set_default("langgate_host", "langgate")

    print(f"Rendering template {args.template} with variables:")
    for key, value in sorted(renderer.variables.items()):
        if "api_key" in key.lower() or "secret" in key.lower():
            print(f"  {key}: [REDACTED]")
        else:
            print(f"  {key}: {value}")

    if DEBUG:
        print("\nRAW ENVIRONMENT VARIABLES:")
        for key, value in sorted(os.environ.items()):
            if "api_key" in key.lower() or "secret" in key.lower():
                print(f"  {key}: [REDACTED]")
            else:
                print(f"  {key}: {value}")

    renderer.render()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Version bumping script for LangGate.

This script automatically updates version numbers across the project:
- Main pyproject.toml
- Package pyproject.toml files
- Helm chart Chart.yaml files

Usage:
    python scripts/bump_version.py [major|minor|patch|--version X.Y.Z]
"""

import argparse
import logging
import re
import sys
import tomllib
from pathlib import Path
from typing import Literal

import yaml

logging.basicConfig(
    level=logging.INFO,
    format="%(message)s",
)
logger = logging.getLogger(__name__)

VersionType = Literal["major", "minor", "patch"]


def parse_version(version: str) -> tuple[int, int, int]:
    """Parse a semantic version string into a tuple of (major, minor, patch)."""
    match = re.match(r"^(\d+)\.(\d+)\.(\d+)$", version)
    if not match:
        raise ValueError(f"Invalid version format: {version}")
    major, minor, patch = map(int, match.groups())
    return (major, minor, patch)


def bump_version(current_version: str, bump_type: VersionType) -> str:
    """Bump the version according to semantic versioning."""
    major, minor, patch = parse_version(current_version)

    if bump_type == "major":
        return f"{major + 1}.0.0"
    if bump_type == "minor":
        return f"{major}.{minor + 1}.0"
    if bump_type == "patch":
        return f"{major}.{minor}.{patch + 1}"
    raise ValueError(f"Invalid bump type: {bump_type}")


def find_pyproject_files(project_root: Path) -> list[Path]:
    """Find all pyproject.toml files in the project."""
    result = [project_root / "pyproject.toml"]

    # Find package pyproject files
    packages_dir = project_root / "packages"
    if packages_dir.exists():
        for package_dir in packages_dir.iterdir():
            if package_dir.is_dir():
                pyproject = package_dir / "pyproject.toml"
                if pyproject.exists():
                    result.append(pyproject)

    return result


def find_chart_files(project_root: Path) -> list[Path]:
    """Find all Chart.yaml files in the project."""
    result = []

    charts_dir = project_root / "deployment" / "k8s" / "charts"
    if charts_dir.exists():
        # Check top-level charts
        for chart_dir in charts_dir.iterdir():
            if chart_dir.is_dir():
                chart_yaml = chart_dir / "Chart.yaml"
                if chart_yaml.exists():
                    result.append(chart_yaml)

        # Check library charts
        library_dir = charts_dir / "library"
        if library_dir.exists() and library_dir.is_dir():
            for lib_chart_dir in library_dir.iterdir():
                if lib_chart_dir.is_dir():
                    chart_yaml = lib_chart_dir / "Chart.yaml"
                    if chart_yaml.exists():
                        result.append(chart_yaml)

    return result


def update_pyproject_toml(file_path: Path, new_version: str) -> None:
    """Update version in a pyproject.toml file."""
    with open(file_path, encoding="utf-8") as f:
        content = f.read()

    pattern = r'(\[project\][^\[]*?\bversion\s*=\s*")[^"]*(")'
    # Use named groups to avoid numeric reference issues
    replacement = r"\g<1>" + new_version + r"\g<2>"
    new_content = re.sub(pattern, replacement, content, flags=re.DOTALL)

    if new_content == content:
        logger.error(f"Could not find project.version in {file_path}")
        return

    # Update langgate package version constraints in dependencies and optional dependencies sections
    # This will find and update langgate-* package versions in (==X.Y.Z) format
    new_content = re.sub(
        r'("langgate-[^"]+)(\s*\(==)[^)]*(\)")',
        r"\g<1>\g<2>" + new_version + r"\g<3>",
        new_content,
    )
    with open(file_path, "w", encoding="utf-8") as f:
        f.write(new_content)

    logger.info(f"Updated {file_path} to version {new_version}")


def update_chart_yaml(file_path: Path, new_version: str) -> None:
    """Update version in a Chart.yaml file using direct text replacement."""
    with open(file_path, encoding="utf-8") as f:
        content = f.read()

    # Update version field - use lambda to avoid backreference issues
    content = re.sub(
        pattern=r"(version:\s*)[\d\.]+",
        repl=lambda m: f"{m.group(1)}{new_version}",
        string=content,
    )

    # Update appVersion field - ensure it's quoted
    content = re.sub(
        pattern=r"(appVersion:\s*)[\"\']?[\d\.]+[\"\']?",
        repl=lambda m: f'{m.group(1)}"{new_version}"',
        string=content,
    )

    # Update dependency versions for langgate components
    content = re.sub(
        pattern=r"(- name: langgate[^\n]+\n\s+repository:[^\n]+\n\s+version:\s*)[\d\.]+",
        repl=lambda m: f"{m.group(1)}{new_version}",
        string=content,
    )

    with open(file_path, "w", encoding="utf-8") as f:
        f.write(content)

    logger.info(f"Updated {file_path} to version {new_version}")


def find_current_version() -> str:
    """Find the current version from the main pyproject.toml file."""
    main_pyproject = Path(__file__).parent.parent / "pyproject.toml"
    if not main_pyproject.exists():
        raise FileNotFoundError(f"Could not find {main_pyproject}")

    with open(main_pyproject, "rb") as f:
        data = tomllib.load(f)

    if "project" in data and "version" in data["project"]:
        return data["project"]["version"]

    raise ValueError("Could not find version in main pyproject.toml")


def update_pyproject_versions(project_root: Path, new_version: str) -> None:
    """Update version in all pyproject.toml files."""
    for pyproject_file in find_pyproject_files(project_root):
        update_pyproject_toml(pyproject_file, new_version)


def update_chart_versions(project_root: Path, new_version: str) -> None:
    """Update version in all Chart.yaml files."""
    for chart_file in find_chart_files(project_root):
        update_chart_yaml(chart_file, new_version)


def update_versions(new_version: str) -> None:
    """Update versions across the project."""
    project_root = Path(__file__).parent.parent
    update_pyproject_versions(project_root, new_version)
    update_chart_versions(project_root, new_version)


def validate_pyproject_versions(project_root: Path, expected_version: str) -> list[str]:
    """Validate versions in pyproject.toml files, returns inconsistencies."""
    inconsistencies = []

    for pyproject_file in find_pyproject_files(project_root):
        with open(pyproject_file, "rb") as f:
            data = tomllib.load(f)

        if (
            "project" in data
            and "version" in data["project"]
            and data["project"]["version"] != expected_version
        ):
            inconsistencies.append(
                f"{pyproject_file}: {data['project']['version']} != {expected_version}"
            )

    return inconsistencies


def validate_chart_versions(project_root: Path, expected_version: str) -> list[str]:
    """Validate versions in Chart.yaml files, returns inconsistencies."""
    inconsistencies = []

    for chart_file in find_chart_files(project_root):
        with open(chart_file, encoding="utf-8") as f:
            data = yaml.safe_load(f)

        if "version" in data and data["version"] != expected_version:
            inconsistencies.append(
                f"{chart_file}: {data['version']} != {expected_version}"
            )

    return inconsistencies


def validate_versions() -> bool:
    """Validate that all versions are consistent."""
    project_root = Path(__file__).parent.parent
    main_version = find_current_version()

    inconsistencies = []
    inconsistencies.extend(validate_pyproject_versions(project_root, main_version))
    inconsistencies.extend(validate_chart_versions(project_root, main_version))

    if inconsistencies:
        logger.error("Version inconsistencies found:")
        for inconsistency in inconsistencies:
            logger.error(f"  - {inconsistency}")
        return False

    logger.info("All versions are consistent!")
    return True


def main() -> None:
    """Run the version bumping script."""
    parser = argparse.ArgumentParser(
        description="Bump version across all LangGate components"
    )
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "bump_type",
        nargs="?",
        choices=["major", "minor", "patch"],
        help="Type of version bump to perform",
    )
    group.add_argument(
        "--version",
        help="Explicitly set version to this value (format: X.Y.Z)",
    )
    group.add_argument(
        "--validate",
        action="store_true",
        help="Validate that all versions are consistent",
    )

    args = parser.parse_args()

    if args.validate:
        sys.exit(0 if validate_versions() else 1)

    current_version = find_current_version()
    new_version = ""

    if args.version:
        # Validate the provided version
        try:
            parse_version(args.version)
            new_version = args.version
        except ValueError as e:
            logger.error(f"Error: {e}")
            sys.exit(1)
    else:
        new_version = bump_version(current_version, args.bump_type)

    logger.info(f"Bumping version from {current_version} to {new_version}")
    update_versions(new_version)
    logger.info(f"\nSuccessfully updated all versions to {new_version}")
    logger.info(
        "\nDon't forget to update CHANGELOG.md with a summary of changes"
        " before creating a release."
    )


if __name__ == "__main__":
    main()



"""Unit tests for the model registry."""

from pathlib import Path
from unittest import mock

from langgate.registry.models import ModelRegistry
from tests.factories.models import LLMInfoFactory
from tests.utils.registry_utils import patch_model_registry


def test_factory_generated_models():
    """Test that factory-generated models are valid."""
    # Create a model with the factory
    model_dict = LLMInfoFactory.create()

    # Validate model fields
    assert isinstance(model_dict, dict)
    assert "id" in model_dict
    assert "name" in model_dict
    assert "provider" in model_dict
    assert "costs" in model_dict
    assert "capabilities" in model_dict
    assert "context_window" in model_dict
    assert "updated_dt" in model_dict


def test_model_registry_works_without_env_file():
    """Test that the model registry works without an environment file."""
    # Create temporary config and models files
    with mock.patch.object(Path, "exists") as mock_exists:
        # Mock the env file to NOT exist
        def side_effect(path):
            # Convert both to strings to avoid Path equality issues
            return ".env" not in str(path)

        mock_exists.side_effect = side_effect

        # Use a mock registry to avoid actual file operations
        with patch_model_registry():
            # Should not raise any exceptions
            registry = ModelRegistry()

            # Verify we have models available
            models = registry.list_llms()
            assert isinstance(models, list)

            # We may have 0 models in this test case since we're using a mock registry
            # The important part is that no exception was raised due to missing .env



from collections.abc import AsyncIterator
from contextlib import asynccontextmanager

from fastapi import FastAPI, Request
from starlette.middleware.cors import CORSMiddleware
from starlette.responses import Response

from langgate.core.logging import configure_logger
from langgate.server.api.api_v1.api import registry_router
from langgate.server.core.config import settings
from langgate.server.core.logging import (
    debug_request,
    get_logger,
    is_debug,
    set_structlog_request_context,
)

configure_logger(json_logs=settings.JSON_LOGS)

logger = get_logger(__name__)
IS_DEBUG = is_debug()


@asynccontextmanager
async def lifespan(app: FastAPI) -> AsyncIterator[None]:
    """Handle application startup and shutdown events."""
    logger.info("application_startup")
    yield
    logger.info("application_shutdown")


app = FastAPI(
    title=settings.PROJECT_NAME,
    openapi_url=f"{settings.API_V1_STR}/openapi.json",
    docs_url=f"{settings.API_V1_STR}/docs",
    redoc_url=f"{settings.API_V1_STR}/redoc",
    lifespan=lifespan,
)

if settings.CORS_ORIGINS:
    app.add_middleware(
        CORSMiddleware,
        allow_origins=[str(origin) for origin in settings.CORS_ORIGINS],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )


@app.middleware("http")
async def logging_middleware(request: Request, call_next) -> Response:
    set_structlog_request_context(request)
    if IS_DEBUG:
        await debug_request(logger, request)
    response: Response = await call_next(request)
    return response


app.include_router(registry_router, prefix=settings.API_V1_STR)



"""Factories for model objects."""

from datetime import UTC
from decimal import Decimal

import factory.declarations as factory
from factory.faker import Faker

from langgate.core.models import (
    MODEL_PROVIDER_ANTHROPIC,
    MODEL_PROVIDER_GOOGLE,
    MODEL_PROVIDER_META,
    MODEL_PROVIDER_OPENAI,
    ContextWindow,
    ImageGenerationCost,
    ImageModelCost,
    ImageModelInfo,
    LLMInfo,
    ModelCapabilities,
    ModelProvider,
    TokenCosts,
)
from tests.factories.base import BasePydanticFactory


class ContextWindowFactory(BasePydanticFactory[ContextWindow]):
    """Factory for ContextWindow objects."""

    max_input_tokens = Faker("pyint", min_value=1024, max_value=200000)
    max_output_tokens = Faker("pyint", min_value=1024, max_value=200000)


class ModelCapabilitiesFactory(BasePydanticFactory[ModelCapabilities]):
    """Factory for ModelCapabilities objects."""

    supports_vision = Faker("pybool")
    supports_tools = Faker("pybool")
    supports_parallel_tool_calls = Faker("pybool")
    supports_response_schema = Faker("pybool")


class TokenCostsFactory(BasePydanticFactory[TokenCosts]):
    """Factory for TokenCosts objects."""

    input_cost_per_token = factory.LazyFunction(lambda: Decimal("0.00001"))
    output_cost_per_token = factory.LazyFunction(lambda: Decimal("0.00002"))


class ModelProviderFactory(BasePydanticFactory[ModelProvider]):
    """Factory for ModelProvider objects."""

    id = factory.Iterator(
        [
            MODEL_PROVIDER_OPENAI,
            MODEL_PROVIDER_ANTHROPIC,
            MODEL_PROVIDER_META,
            MODEL_PROVIDER_GOOGLE,
        ]
    )
    name = factory.LazyAttribute(lambda o: o.id.title())
    description = Faker("sentence")


class LLMInfoFactory(BasePydanticFactory[LLMInfo]):
    """Factory for LLMInfo objects."""

    id = factory.Sequence(lambda n: f"model-{n}")
    name = factory.Sequence(lambda n: f"Test Model {n}")
    description = Faker("sentence")
    provider = factory.SubFactory(ModelProviderFactory)
    costs = factory.SubFactory(TokenCostsFactory)
    capabilities = factory.SubFactory(ModelCapabilitiesFactory)
    context_window = factory.SubFactory(ContextWindowFactory)
    updated_dt = Faker("date_time_this_year", tzinfo=UTC)


class ImageGenerationCostFactory(BasePydanticFactory[ImageGenerationCost]):
    """Factory for ImageGenerationCost objects."""

    input_cost_per_image = factory.LazyFunction(lambda: Decimal("0.04"))
    width = Faker("pyint", min_value=512, max_value=2048)
    height = Faker("pyint", min_value=512, max_value=2048)


class ImageModelCostFactory(BasePydanticFactory[ImageModelCost]):
    """Factory for ImageModelCost objects."""

    flat_rate = factory.SubFactory(ImageGenerationCostFactory)


class ImageModelInfoFactory(BasePydanticFactory[ImageModelInfo]):
    """Factory for ImageModelInfo objects."""

    id = factory.Sequence(lambda n: f"image-model-{n}")
    name = factory.Sequence(lambda n: f"Test Image Model {n}")
    description = Faker("sentence")
    provider = factory.SubFactory(ModelProviderFactory)
    costs = factory.SubFactory(ImageModelCostFactory)
    updated_dt = Faker("date_time_this_year", tzinfo=UTC)



"""Test metadata override functionality for models in YAML config."""

import tempfile
from pathlib import Path

from langgate.registry.config import RegistryConfig
from langgate.registry.models import ModelRegistry


def test_yaml_config_overrides_json_metadata_for_llm():
    """Test that metadata in YAML config overrides JSON data for LLM models."""
    # Create a temporary JSON file with model data
    models_json = {
        "gemini/gemini-2.5-pro": {
            "name": "Gemini 2.5 Pro",
            "mode": "text",
            "model_provider": "google",
            "model_provider_name": "Google",
            "capabilities": {
                "supports_tools": True,
                "supports_vision": True,
                "supports_reasoning": True,
                "supports_response_schema": True,
                "supports_system_messages": True,
                "supports_tool_choice": True,
            },
            "context": {"max_input_tokens": 1048576, "max_output_tokens": 65536},
            "costs": {
                "input_cost_per_token": "0.00000125",
                "output_cost_per_token": "0.00001",
            },
            "description": "Gemini 2.5 Pro is Google's state-of-the-art thinking model.",
        }
    }

    # Create a temporary YAML config file with overrides
    yaml_config = """
default_params:
  temperature: 0.7

services:
  gemini:
    api_key: "${GEMINI_API_KEY}"
    base_url: "https://api.gemini.ai"

models:
  text:
    - id: google/gemini-no-reasoning
      service:
        provider: gemini
        model_id: gemini-2.5-pro
      name: Gemini no reasoning
      description: Custom Gemini without reasoning capability
      capabilities:
        supports_tools: true
        supports_vision: true
        supports_response_schema: true
        supports_system_messages: true
        supports_tool_choice: true
        supports_reasoning: false  # Override: disable reasoning
      context:
        max_input_tokens: 100000  # Override: reduce context
        max_output_tokens: 65000  # Override: reduce output
      costs:
        input_cost_per_token: "0.00000200"  # Override: different pricing
        output_cost_per_token: "0.00001500"
      override_params:
        temperature: 0.2
"""

    with tempfile.NamedTemporaryFile(mode="w", suffix=".json", delete=False) as f:
        import json

        json.dump(models_json, f)
        models_path = Path(f.name)

    with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
        f.write(yaml_config)
        config_path = Path(f.name)

    try:
        # Create registry with test files
        config = RegistryConfig(
            models_data_path=models_path,
            config_path=config_path,
        )
        registry = ModelRegistry(config)

        # Get the model info
        model = registry.get_llm_info("google/gemini-no-reasoning")

        # Assert that YAML overrides are applied
        assert model.name == "Gemini no reasoning"
        assert model.description == "Custom Gemini without reasoning capability"

        # Check capabilities override
        assert model.capabilities.supports_reasoning is False  # Overridden
        assert model.capabilities.supports_tools is True
        assert model.capabilities.supports_vision is True

        # Check context override
        assert model.context_window.max_input_tokens == 100000  # Overridden
        assert model.context_window.max_output_tokens == 65000  # Overridden

        # Check costs override
        assert str(model.costs.input_cost_per_token) == "0.00000200"  # Overridden
        assert str(model.costs.output_cost_per_token) == "0.00001500"  # Overridden

    finally:
        models_path.unlink(missing_ok=True)
        config_path.unlink(missing_ok=True)


def test_yaml_config_partial_override_preserves_json_defaults():
    """Test that partial overrides in YAML preserve non-overridden JSON data."""
    # Create a temporary JSON file with model data
    models_json = {
        "openai/gpt-5": {
            "name": "GPT-5",
            "mode": "text",
            "model_provider": "openai",
            "model_provider_name": "OpenAI",
            "capabilities": {
                "supports_tools": True,
                "supports_vision": True,
                "supports_reasoning": True,
            },
            "context": {"max_input_tokens": 200000, "max_output_tokens": 4096},
            "costs": {
                "input_cost_per_token": "0.00001",
                "output_cost_per_token": "0.00002",
            },
            "description": "OpenAI's GPT-5 model.",
        }
    }

    # Create a YAML config with partial override (only capabilities)
    yaml_config = """
services:
  openai:
    api_key: "${OPENAI_API_KEY}"
    base_url: "https://api.openai.com/v1"

models:
  text:
    - id: openai/gpt-5-no-vision
      service:
        provider: openai
        model_id: gpt-5
      capabilities:
        supports_vision: false  # Only override vision capability
"""

    with tempfile.NamedTemporaryFile(mode="w", suffix=".json", delete=False) as f:
        import json

        json.dump(models_json, f)
        models_path = Path(f.name)

    with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
        f.write(yaml_config)
        config_path = Path(f.name)

    try:
        config = RegistryConfig(
            models_data_path=models_path,
            config_path=config_path,
        )
        registry = ModelRegistry(config)

        model = registry.get_llm_info("openai/gpt-5-no-vision")

        # Original values should be preserved
        assert model.name == "GPT-5"  # From JSON
        assert model.description == "OpenAI's GPT-5 model."  # From JSON
        assert model.context_window.max_input_tokens == 200000  # From JSON
        assert model.context_window.max_output_tokens == 4096  # From JSON
        assert str(model.costs.input_cost_per_token) == "0.00001"  # From JSON

        # Overridden values should be applied
        assert model.capabilities.supports_vision is False  # Overridden
        assert model.capabilities.supports_tools is True  # From JSON
        assert model.capabilities.supports_reasoning is True  # From JSON

    finally:
        models_path.unlink(missing_ok=True)
        config_path.unlink(missing_ok=True)


def test_yaml_config_overrides_for_image_models():
    """Test that metadata overrides work for image generation models."""
    models_json = {
        "openai/dall-e-3": {
            "name": "DALL-E 3",
            "mode": "image",
            "model_provider": "openai",
            "model_provider_name": "OpenAI",
            "costs": {
                "image_generation": {
                    "quality_tiers": {
                        "standard": {
                            "1024x1024": "0.040",
                            "1024x1792": "0.080",
                        },
                        "hd": {
                            "1024x1024": "0.080",
                            "1024x1792": "0.120",
                        },
                    }
                }
            },
            "description": "DALL-E 3 image generation model.",
        }
    }

    yaml_config = """
services:
  openai:
    api_key: "${OPENAI_API_KEY}"
    base_url: "https://api.openai.com/v1"

models:
  image:
    - id: openai/dall-e-3-budget
      service:
        provider: openai
        model_id: dall-e-3
      name: DALL-E 3 Budget
      description: Budget version with reduced pricing
      costs:
        image_generation:
          flat_rate: "0.020"  # Override with flat rate pricing
"""

    with tempfile.NamedTemporaryFile(mode="w", suffix=".json", delete=False) as f:
        import json

        json.dump(models_json, f)
        models_path = Path(f.name)

    with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
        f.write(yaml_config)
        config_path = Path(f.name)

    try:
        config = RegistryConfig(
            models_data_path=models_path,
            config_path=config_path,
        )
        registry = ModelRegistry(config)

        model = registry.get_image_model_info("openai/dall-e-3-budget")

        assert model.name == "DALL-E 3 Budget"
        assert model.description == "Budget version with reduced pricing"
        assert str(model.costs.image_generation.flat_rate) == "0.020"
        assert (
            model.costs.image_generation.quality_tiers is None
        )  # Replaced with flat rate

    finally:
        models_path.unlink(missing_ok=True)
        config_path.unlink(missing_ok=True)



"""Tests for environment configuration."""

import os
from pathlib import Path
from unittest import mock
from unittest.mock import MagicMock

import pytest
import yaml
from pydantic_settings import BaseSettings, SettingsConfigDict

from langgate.server.core.config import (
    ApiSettings,
    FixedYamlConfigSettingsSource,
    _get_api_env_file_path,
)
from tests.utils.config_utils import config_path_resolver
from tests.utils.server_utils import prevent_server_config_loading
from tests.utils.utils import capture_logs


def test_app_config_loading_from_yaml(
    mock_config_path_in_env: dict[str, str], mock_env_vars_from_env_file: dict[str, str]
):
    """Test that app config is correctly loaded from YAML file and logs are emitted."""
    with capture_logs() as caplogs:
        # Re-initialize settings to ensure fresh loading
        test_settings = ApiSettings()

        # Test that settings were properly loaded from the YAML file
        assert test_settings.PROJECT_NAME == "LangGate Test"
        assert test_settings.CORS_ORIGINS == ["http://localhost"]
        assert test_settings.HTTPS is False
        assert test_settings.SECRET_KEY.get_secret_value() == "test-secret-key"

        # Verify the log was emitted
        assert any(log["event"] == "loaded_app_config" for log in caplogs)


def test_warning_when_no_app_config(tmp_path: Path):
    """Test warning when YAML has no app_config section."""
    # Create a config file without app_config
    config_path = tmp_path / "no_app_config.yaml"
    with open(config_path, "w") as f:
        yaml.dump({"some_other_key": "value"}, f)

    class SimpleSettings(BaseSettings):
        """Simple settings class for testing FixedYamlConfigSettingsSource."""

    with (
        mock.patch.dict(os.environ, {"LANGGATE_CONFIG": str(config_path)}),
        capture_logs() as caplogs,
    ):
        yaml_source = FixedYamlConfigSettingsSource(SimpleSettings)
        # Call it to trigger config loading
        yaml_source()

        # Check warning was logged
        assert any(log["event"] == "no_app_config_in_config_file" for log in caplogs)


def test_fixed_yaml_config_env_path():
    """Test that LANGGATE_CONFIG environment variable is used by FixedYamlConfigSettingsSource."""
    with mock.patch.dict(
        os.environ, {"LANGGATE_CONFIG": "/custom/path/langgate_config.yaml"}
    ):

        class TestSettings(BaseSettings):
            pass

        yaml_source = FixedYamlConfigSettingsSource(TestSettings)
        assert str(yaml_source.config_path) == "/custom/path/langgate_config.yaml"


@pytest.mark.parametrize(
    "source,expected_path",
    [
        # Skip arg path as FixedYamlConfigSettingsSource doesn't accept constructor args
        ("env", "/env/path/config.yaml"),
        ("cwd", "langgate_config.yaml"),
        ("package_dir", "default_config.yaml"),
    ],
    ids=["env_var", "cwd_path", "package_dir_path"],
)
def test_fixed_yaml_config_settings_source_paths(source, expected_path):
    """Test path resolution in FixedYamlConfigSettingsSource with different sources."""

    class TestSettings(BaseSettings):
        """Simple settings class for testing FixedYamlConfigSettingsSource."""

    with (
        prevent_server_config_loading(),
        config_path_resolver(source, "config_yaml", expected_path),
    ):
        # Initialize the YAML settings source
        yaml_source = FixedYamlConfigSettingsSource(TestSettings)

        # Verify the config path was correctly resolved
        assert expected_path in str(yaml_source.config_path)


def test_api_settings_customise_sources():
    """Test that ApiSettings customizes sources by adding FixedYamlConfigSettingsSource."""
    # Directly test the behavior of settings_customise_sources

    # Create mock settings sources
    init_settings = MagicMock()
    env_settings = MagicMock()
    dotenv_settings = MagicMock()
    file_secret_settings = MagicMock()

    settings_args = [init_settings, env_settings, dotenv_settings, file_secret_settings]

    # Get the actual settings sources
    sources = ApiSettings.settings_customise_sources(
        ApiSettings,
        init_settings,
        env_settings,
        dotenv_settings,
        file_secret_settings,
    )

    # Verify that we have settings sources (should be 5)
    assert len(sources) > 0

    # Verify that we're returning all the provided sources
    for arg in settings_args:
        assert arg in sources, f"{arg.__name__} should be included in sources"

    # Verify that FixedYamlConfigSettingsSource is included
    assert any(
        isinstance(source, FixedYamlConfigSettingsSource) for source in sources
    ), "FixedYamlConfigSettingsSource should be included in sources"

    # Verify the order of sources
    for i, (arg, source) in enumerate(zip(settings_args, sources, strict=False)):
        assert arg == source, f"""Source at index {i} mismatch:
         expected {arg.__class__.__name__}, got {source.__class__.__name__}"""

    assert isinstance(sources[-1], FixedYamlConfigSettingsSource), (
        "Fifth source should be FixedYamlConfigSettingsSource"
    )


def test_api_settings_env_file_source_resolution():
    """Test that ApiSettings uses correct env file resolution based on source."""

    # Since ApiSettings.model_config is evaluated at import time, we need to test that
    # the _get_api_env_file_path function returns the correct path, and then
    # verify that this path is what would be used in the model_config

    def _set_settings_config(env_file: Path | str | None):
        return SettingsConfigDict(
            case_sensitive=True,
            env_file=env_file,
            env_file_encoding="utf-8",
            extra="ignore",
        )

    # Test case 1: Using environment variable LANGGATE_ENV_FILE
    expected_path = "/custom/env/path/.env"
    with config_path_resolver("env", "env_file", expected_path):
        env_file_path = _get_api_env_file_path()
        assert str(env_file_path) == expected_path

        # Verify this path would be used in settings config
        config_dict = _set_settings_config(env_file_path)
        assert str(config_dict.get("env_file")) == expected_path

    # Test case 2: Using CWD when no LANGGATE_ENV_FILE is set
    expected_path = "/fake/cwd/.env"
    with config_path_resolver("cwd", "env_file", expected_path):
        env_file_path = _get_api_env_file_path()
        assert str(env_file_path) == expected_path

        config_dict = _set_settings_config(env_file_path)
        assert str(config_dict.get("env_file")) == expected_path

    # Test case 3: No LANGGATE_ENV_FILE env var and no env file in CWD
    with mock.patch("pathlib.Path.exists", return_value=False):
        env_file_path = _get_api_env_file_path()
        assert env_file_path is None

        # Verify None would be used in settings config
        config_dict = _set_settings_config(env_file_path)
        assert config_dict.get("env_file") is None



"""Integration tests for LocalRegistryClient."""

from datetime import timedelta

import pytest

from langgate.core.models import ImageModelInfo, LLMInfo
from langgate.registry.local import LocalRegistryClient
from tests.mocks.registry_mocks import (
    CustomImageModelInfo,
    CustomLLMInfo,
    CustomLocalRegistryClient,
)


@pytest.mark.asyncio
async def test_local_registry_client_get_model(
    local_registry_client: LocalRegistryClient,
):
    """Test getting a model from the registry via the client."""
    # First list all models
    models = await local_registry_client.list_llms()
    assert len(models) > 0

    # Pick the first model
    first_model = models[0]

    # Then get that specific model by ID
    model = await local_registry_client.get_llm_info(first_model.id)

    assert model.id == first_model.id
    assert model.name == first_model.name
    assert isinstance(model, LLMInfo)
    assert model.provider is not None


@pytest.mark.asyncio
async def test_local_registry_client_list_models(
    local_registry_client: LocalRegistryClient,
):
    """Test listing all models from the registry."""
    models = await local_registry_client.list_llms()

    assert len(models) > 0

    for model in models:
        assert isinstance(model, LLMInfo)
        assert model.id is not None
        assert model.name is not None
        assert model.provider is not None
        assert model.costs is not None


@pytest.mark.asyncio
async def test_local_registry_client_caching(
    local_registry_client: LocalRegistryClient,
):
    """Test that models are properly cached.

    Note that caching at the client level is intended for
    HTTP clients. For simplicity, this logic lives in the base class
    `BaseRegistryClient` but refreshing the cache is essentially
    a no-op for the local registry client.
    """
    last_cache_refresh = local_registry_client._last_cache_refresh
    assert last_cache_refresh is None

    # Initial request populates cache
    models = await local_registry_client.list_llms()
    assert len(models) > 0

    # Verify cache state
    assert local_registry_client._last_cache_refresh is not None
    last_cache_refresh = local_registry_client._last_cache_refresh
    assert len(local_registry_client._llm_cache) > 0

    # Get a specific model ID to test
    model_id = models[0].id

    # This should use the cache
    model = await local_registry_client.get_llm_info(model_id)

    assert model.id == model_id

    # Verify it's the same object reference (from cache)
    assert model is local_registry_client._llm_cache[model_id]
    assert last_cache_refresh == local_registry_client._last_cache_refresh

    # Fetch the same model again
    model2 = await local_registry_client.get_llm_info(model_id)
    assert model2.id == model_id
    assert model2 is model

    # Simulate cache expiration
    expired_last_refresh = (
        local_registry_client._last_cache_refresh
        - local_registry_client._cache_ttl
        - timedelta(seconds=1)
    )
    local_registry_client._last_cache_refresh = expired_last_refresh

    # Fetch the model again, which should refresh the cache
    model3 = await local_registry_client.get_llm_info(model_id)
    assert model3.id == model_id
    # for the default local registry client, we expect refreshing to still return the same objects
    # which are perpetually cached in the ModelRegistry.
    assert model3 is model
    assert model3 is local_registry_client._llm_cache[model_id]

    # Verify cache state
    assert local_registry_client._last_cache_refresh > expired_last_refresh
    assert local_registry_client._last_cache_refresh > last_cache_refresh


@pytest.mark.asyncio
async def test_custom_local_registry_client(
    custom_local_registry_client: CustomLocalRegistryClient,
):
    """Test using a custom client with a custom schema."""
    # List models with the custom client
    models = await custom_local_registry_client.list_llms()
    assert len(models) > 0

    # Verify custom model type
    for model in models:
        assert isinstance(model, CustomLLMInfo)
        assert model.custom_field == "custom_value"

    # Get a specific model
    first_model = models[0]
    model = await custom_local_registry_client.get_llm_info(first_model.id)

    # Verify it's the correct custom type
    assert isinstance(model, CustomLLMInfo)
    assert model.id == first_model.id
    assert model.name == first_model.name
    assert model.custom_field == "custom_value"


@pytest.mark.asyncio
async def test_custom_local_registry_client_caching(
    custom_local_registry_client: CustomLocalRegistryClient,
):
    """Test that models are properly cached."""
    last_cache_refresh = custom_local_registry_client._last_cache_refresh
    assert last_cache_refresh is None

    # Initial request populates cache
    models = await custom_local_registry_client.list_llms()
    assert len(models) > 0

    # Verify cache state
    assert custom_local_registry_client._last_cache_refresh is not None
    last_cache_refresh = custom_local_registry_client._last_cache_refresh
    assert len(custom_local_registry_client._llm_cache) > 0

    # Get a specific model ID to test
    model_id = models[0].id

    # This should use the cache
    model = await custom_local_registry_client.get_llm_info(model_id)

    assert model.id == model_id

    # Verify it's the same object reference (from cache)
    assert model is custom_local_registry_client._llm_cache[model_id]
    assert last_cache_refresh == custom_local_registry_client._last_cache_refresh

    # Fetch the same model again
    model2 = await custom_local_registry_client.get_llm_info(model_id)
    assert model2.id == model_id
    assert model2 is model

    # Simulate cache expiration
    expired_last_refresh = (
        custom_local_registry_client._last_cache_refresh
        - custom_local_registry_client._cache_ttl
        - timedelta(seconds=1)
    )
    custom_local_registry_client._last_cache_refresh = expired_last_refresh

    # Fetch the model again, which should refresh the cache
    model3 = await custom_local_registry_client.get_llm_info(model_id)
    assert model3.id == model_id
    # for a custom local registry client, we expect refreshing to return new objects
    # as they will be revalidated against the custom schema.
    assert model3 is not model
    assert model3 is custom_local_registry_client._llm_cache[model_id]

    # Verify cache state
    assert custom_local_registry_client._last_cache_refresh > expired_last_refresh
    assert custom_local_registry_client._last_cache_refresh > last_cache_refresh


# Image model tests


@pytest.mark.asyncio
async def test_local_registry_client_get_image_model(
    local_registry_client: LocalRegistryClient,
):
    """Test getting an image model from the registry via the client."""
    # First list all image models
    models = await local_registry_client.list_image_models()
    assert len(models) > 0

    # Pick the first model
    first_model = models[0]

    # Then get that specific model by ID
    model = await local_registry_client.get_image_model_info(first_model.id)

    assert model.id == first_model.id
    assert model.name == first_model.name
    assert isinstance(model, ImageModelInfo)
    assert model.provider is not None
    assert model.costs is not None
    assert model.costs.image_generation is not None


@pytest.mark.asyncio
async def test_local_registry_client_list_image_models(
    local_registry_client: LocalRegistryClient,
):
    """Test listing all image models from the registry."""
    models = await local_registry_client.list_image_models()

    assert len(models) > 0

    for model in models:
        assert isinstance(model, ImageModelInfo)
        assert model.id is not None
        assert model.name is not None
        assert model.provider is not None
        assert model.costs is not None
        assert model.costs.image_generation is not None


@pytest.mark.asyncio
async def test_local_registry_client_image_model_caching(
    local_registry_client: LocalRegistryClient,
):
    """Test that image models are properly cached."""
    last_cache_refresh = local_registry_client._last_cache_refresh
    assert last_cache_refresh is None

    # Initial request populates cache
    models = await local_registry_client.list_image_models()
    assert len(models) > 0

    # Verify cache state
    assert local_registry_client._last_cache_refresh is not None
    last_cache_refresh = local_registry_client._last_cache_refresh
    assert len(local_registry_client._image_cache) > 0

    # Get a specific model ID to test
    model_id = models[0].id

    # This should use the cache
    model = await local_registry_client.get_image_model_info(model_id)

    assert model.id == model_id

    # Verify it's the same object reference (from cache)
    assert model is local_registry_client._image_cache[model_id]
    assert last_cache_refresh == local_registry_client._last_cache_refresh

    # Fetch the same model again
    model2 = await local_registry_client.get_image_model_info(model_id)
    assert model2.id == model_id
    assert model2 is model

    # Simulate cache expiration
    expired_last_refresh = (
        local_registry_client._last_cache_refresh
        - local_registry_client._cache_ttl
        - timedelta(seconds=1)
    )
    local_registry_client._last_cache_refresh = expired_last_refresh

    # Fetch the model again, which should refresh the cache
    model3 = await local_registry_client.get_image_model_info(model_id)
    assert model3.id == model_id
    # for the default local registry client, we expect refreshing to still return the same objects
    # which are perpetually cached in the ModelRegistry.
    assert model3 is model
    assert model3 is local_registry_client._image_cache[model_id]

    # Verify cache state
    assert local_registry_client._last_cache_refresh > expired_last_refresh
    assert local_registry_client._last_cache_refresh > last_cache_refresh


@pytest.mark.asyncio
async def test_custom_local_registry_client_image_models(
    custom_local_registry_client: CustomLocalRegistryClient,
):
    """Test using a custom client with custom image model schema."""
    # List image models with the custom client
    models = await custom_local_registry_client.list_image_models()
    assert len(models) > 0

    # Verify custom model type
    for model in models:
        assert isinstance(model, CustomImageModelInfo)
        assert model.custom_field == "custom_value"

    # Get a specific model
    first_model = models[0]
    model = await custom_local_registry_client.get_image_model_info(first_model.id)

    # Verify it's the correct custom type
    assert isinstance(model, CustomImageModelInfo)
    assert model.id == first_model.id
    assert model.name == first_model.name
    assert model.custom_field == "custom_value"


@pytest.mark.asyncio
async def test_custom_local_registry_client_image_model_caching(
    custom_local_registry_client: CustomLocalRegistryClient,
):
    """Test that image models are properly cached in custom client."""
    last_cache_refresh = custom_local_registry_client._last_cache_refresh
    assert last_cache_refresh is None

    # Initial request populates cache
    models = await custom_local_registry_client.list_image_models()
    assert len(models) > 0

    # Verify cache state
    assert custom_local_registry_client._last_cache_refresh is not None
    last_cache_refresh = custom_local_registry_client._last_cache_refresh
    assert len(custom_local_registry_client._image_cache) > 0

    # Get a specific model ID to test
    model_id = models[0].id

    # This should use the cache
    model = await custom_local_registry_client.get_image_model_info(model_id)

    assert model.id == model_id

    # Verify it's the same object reference (from cache)
    assert model is custom_local_registry_client._image_cache[model_id]
    assert last_cache_refresh == custom_local_registry_client._last_cache_refresh

    # Fetch the same model again
    model2 = await custom_local_registry_client.get_image_model_info(model_id)
    assert model2.id == model_id
    assert model2 is model

    # Simulate cache expiration
    expired_last_refresh = (
        custom_local_registry_client._last_cache_refresh
        - custom_local_registry_client._cache_ttl
        - timedelta(seconds=1)
    )
    custom_local_registry_client._last_cache_refresh = expired_last_refresh

    # Fetch the model again, which should refresh the cache
    model3 = await custom_local_registry_client.get_image_model_info(model_id)
    assert model3.id == model_id
    # for a custom local registry client, we expect refreshing to return new objects
    # as they will be revalidated against the custom schema.
    assert model3 is not model
    assert model3 is custom_local_registry_client._image_cache[model_id]

    # Verify cache state
    assert custom_local_registry_client._last_cache_refresh > expired_last_refresh
    assert custom_local_registry_client._last_cache_refresh > last_cache_refresh



"""Protocol definitions for parameter transformers."""

from typing import Any, Protocol

from langgate.core.logging import get_logger

logger = get_logger(__name__)


class TransformerClientProtocol(Protocol):
    """Protocol for transformer clients."""

    async def get_params(
        self, model_id: str, input_params: dict[str, Any]
    ) -> tuple[str, dict[str, Any]]:
        """Get transformed parameters for the specified model.

        Args:
            model_id: The ID of the model to get transformed parameters for
            input_params: The parameters to transform

        Returns:
            A tuple containing (api_format, transformed_parameters)
        """
        ...



"""Client API fixtures."""

from collections.abc import AsyncGenerator, Generator

import pytest
import pytest_asyncio
from httpx import ASGITransport, AsyncClient

from langgate.client.http import HTTPRegistryClient
from langgate.server.core.config import settings
from langgate.server.main import app
from tests.mocks.client_mocks import CustomHTTPRegistryClient


@pytest_asyncio.fixture
async def registry_api_client() -> AsyncGenerator[AsyncClient]:
    """Return an async client for testing the registry API."""
    # Create ASGITransport for testing
    transport = ASGITransport(app=app)

    async with AsyncClient(
        transport=transport,
        base_url=f"http://{settings.TEST_SERVER_HOST}{settings.API_V1_STR}",
    ) as client:
        yield client


@pytest_asyncio.fixture
async def http_registry_client(
    registry_api_client,
) -> AsyncGenerator[HTTPRegistryClient]:
    """Return an HTTPRegistryClient configured to use the test server."""
    # Save original singleton state
    original_instance = HTTPRegistryClient._instance
    HTTPRegistryClient._instance = None

    client = HTTPRegistryClient(
        base_url=f"http://{settings.TEST_SERVER_HOST}{settings.API_V1_STR}",
        api_key=None,
        http_client=registry_api_client,
    )

    yield client

    # Restore original singleton state
    HTTPRegistryClient._instance = original_instance


@pytest.fixture
def custom_http_registry_client(
    registry_api_client,
) -> Generator[CustomHTTPRegistryClient]:
    """Return a CustomHTTPRegistryClient instance for testing custom schema handling."""
    client = CustomHTTPRegistryClient(
        base_url=f"http://{settings.TEST_SERVER_HOST}{settings.API_V1_STR}",
        api_key=None,
        http_client=registry_api_client,
    )

    yield client



import structlog
from fastapi import HTTPException, status

from langgate.registry import ModelRegistry
from langgate.registry.models import ImageModelInfo, LLMInfo

logger = structlog.stdlib.get_logger(__name__)


class ModelRegistryAPI:
    """
    API for accessing the model registry.
    This will be expanded when gateway becomes a microservice.
    """

    def __init__(self, registry: ModelRegistry | None = None):
        self.registry = registry or ModelRegistry()

    async def get_llm_info(self, model_id: str) -> LLMInfo:
        """Get LLM information."""
        try:
            return self.registry.get_llm_info(model_id)
        except ValueError as exc:
            if "not found" in str(exc):
                raise HTTPException(
                    status_code=status.HTTP_404_NOT_FOUND,
                    detail=f"Model {model_id} not found",
                ) from exc
            await logger.aexception("model_info_error", model_id=model_id)
            raise

    async def list_llms(self) -> list[LLMInfo]:
        """List all available LLMs."""
        return self.registry.list_llms()

    async def get_image_model_info(self, model_id: str) -> ImageModelInfo:
        """Get image model information."""
        try:
            return self.registry.get_image_model_info(model_id)
        except ValueError as exc:
            if "not found" in str(exc):
                raise HTTPException(
                    status_code=status.HTTP_404_NOT_FOUND,
                    detail=f"Model {model_id} not found",
                ) from exc
            await logger.aexception("model_info_error", model_id=model_id)
            raise

    async def list_image_models(self) -> list[ImageModelInfo]:
        """List all available image models."""
        return self.registry.list_image_models()



"""Integration tests for LocalTransformerClient."""

import os
from pathlib import Path
from typing import Literal
from unittest import mock

import pytest
from pydantic.types import SecretStr

from langgate.core.schemas.config import ConfigSchema
from langgate.transform.local import LocalTransformerClient
from tests.utils.config_utils import config_path_resolver, patch_load_yaml_config


@pytest.mark.parametrize(
    "source,expected_path",
    [
        ("arg", "/arg/path/langgate_config.yaml"),
        ("env", "/env/path/langgate_config.yaml"),
        ("cwd", "langgate_config.yaml"),
        ("package_dir", "default_config.yaml"),
    ],
    ids=["arg_path", "env_var", "cwd_path", "package_dir_path"],
)
def test_transformer_config_yaml_paths(
    source: Literal["arg", "env", "cwd", "package_dir"], expected_path: str
):
    """Test path resolution for config YAML file with different sources."""
    # Reset singleton for each case
    LocalTransformerClient._instance = None

    with config_path_resolver(source, "config_yaml", expected_path):
        # For arg source, we need to explicitly pass the path
        if source == "arg":
            client = LocalTransformerClient(config_path=Path(expected_path))
        else:
            client = LocalTransformerClient()

        assert expected_path in str(client.config_path)


@pytest.mark.parametrize(
    "source,expected_path",
    [
        ("arg", "/arg/path/.env"),
        ("env", "/env/path/.env"),
        ("cwd", ".env"),
    ],
    ids=["arg_path", "env_var", "cwd_path"],
)
def test_transformer_env_file_paths(source, expected_path):
    """Test path resolution for .env file with different sources."""
    # Reset singleton for each case
    LocalTransformerClient._instance = None

    with config_path_resolver(source, "env_file", expected_path):
        # For arg source, we need to explicitly pass the path
        if source == "arg":
            client = LocalTransformerClient(env_file_path=Path(expected_path))
        else:
            client = LocalTransformerClient()

        assert expected_path in str(client.env_file_path)


@pytest.mark.asyncio
async def test_transformer_config_loading(
    local_transformer_client: LocalTransformerClient,
):
    """Test that configuration is properly loaded."""
    # Check service configs are loaded
    assert "openai" in local_transformer_client._service_config
    assert "anthropic" in local_transformer_client._service_config

    # Check global defaults are loaded (now modality-specific)
    assert "text" in local_transformer_client._global_config["default_params"]
    assert (
        "temperature"
        in local_transformer_client._global_config["default_params"]["text"]
    )

    # Check model mappings are processed
    assert "gpt-4o" in local_transformer_client._model_mappings
    assert (
        "anthropic/claude-sonnet-4-reasoning"
        in local_transformer_client._model_mappings
    )


@pytest.mark.asyncio
async def test_transformer_global_defaults(
    local_transformer_client: LocalTransformerClient,
):
    """Test applying global default parameters."""
    # Test with empty params
    _, result = await local_transformer_client.get_params("gpt-4o", {})

    # Global default should be applied
    assert result["temperature"] == 0.7


@pytest.mark.asyncio
async def test_transformer_service_provider_defaults(
    local_transformer_client: LocalTransformerClient,
):
    """Test applying service provider default parameters."""
    # OpenAI service provider defaults
    _, result = await local_transformer_client.get_params("gpt-4o", {})
    assert result["max_tokens"] == 1000

    # Anthropic service provider defaults
    _, result = await local_transformer_client.get_params(
        "anthropic/claude-sonnet-4", {}
    )
    assert result["max_tokens"] == 2000


@pytest.mark.asyncio
async def test_transformer_model_pattern_params(
    local_transformer_client: LocalTransformerClient,
):
    """Test applying model pattern specific parameters."""
    # Anthropic reasoning pattern
    _, result = await local_transformer_client.get_params(
        "anthropic/claude-sonnet-4-reasoning", {}
    )

    # Pattern should apply thinking parameter and remove temperature
    assert result["thinking"]["type"] == "enabled"
    assert "temperature" not in result

    # Pattern should apply default max_tokens from model_patterns
    assert result["max_tokens"] == 64000


@pytest.mark.asyncio
async def test_transformer_model_pattern_defaults(
    local_transformer_client: LocalTransformerClient,
):
    """Test applying default parameters at the model pattern level."""
    # Use empty params to ensure defaults are applied
    _, result = await local_transformer_client.get_params(
        "anthropic/claude-sonnet-4-reasoning", {}
    )

    # The pattern has default_params with max_tokens
    assert result["max_tokens"] == 64000

    # User params should still override pattern defaults
    user_params = {"max_tokens": 1000}
    _, result = await local_transformer_client.get_params(
        "anthropic/claude-sonnet-4-reasoning", user_params
    )
    assert result["max_tokens"] == 1000


@pytest.mark.asyncio
async def test_transformer_model_pattern_renames(
    local_transformer_client: LocalTransformerClient,
):
    """Test parameter renaming at the model pattern level."""
    # Anthropic reasoning pattern has reasoning -> thinking rename
    user_params = {"reasoning": {"depth": "deep"}, "temperature": 0.7}
    _, result = await local_transformer_client.get_params(
        "anthropic/claude-sonnet-4-reasoning", user_params
    )

    # Check parameter was renamed via pattern's rename_params
    assert "reasoning" not in result
    assert "thinking" in result
    assert result["thinking"]["depth"] == "deep"
    # Temperature should be removed by pattern's remove_params
    assert "temperature" not in result


async def test_transformer_modality_specific_service_defaults():
    """Test that service-level default_params are applied by modality."""
    LocalTransformerClient._instance = None

    custom_config = {
        "default_params": {
            "text": {"global_temperature": 0.7},
            "image": {"global_size": "1024x1024"},
        },
        "services": {
            "test_service": {
                "api_key": "${TEST_API_KEY}",
                "base_url": "https://api.test.com/v1",
                "default_params": {
                    "text": {
                        "max_tokens": 2000,
                        "stream_usage": True,
                    },
                    "image": {
                        "quality": "hd",
                        "style": "vivid",
                    },
                },
            },
        },
        "models": {
            "text": [
                {
                    "id": "test_service/text-model",
                    "service": {
                        "provider": "test_service",
                        "model_id": "actual-text-model",
                    },
                }
            ],
            "image": [
                {
                    "id": "test_service/image-model",
                    "service": {
                        "provider": "test_service",
                        "model_id": "actual-image-model",
                    },
                }
            ],
        },
        "app_config": {},
    }

    with (
        mock.patch.dict(os.environ, {"TEST_API_KEY": "test-key"}),
        patch_load_yaml_config(ConfigSchema.model_validate(custom_config)),
    ):
        client = LocalTransformerClient()

        # Test text model gets text-specific service defaults
        _, text_result = await client.get_params("test_service/text-model", {})
        assert text_result["max_tokens"] == 2000
        assert text_result["stream_usage"] is True
        assert text_result["global_temperature"] == 0.7
        # Should not have image-specific defaults
        assert "quality" not in text_result
        assert "style" not in text_result
        assert "global_size" not in text_result

        # Test image model gets image-specific service defaults
        _, image_result = await client.get_params("test_service/image-model", {})
        assert image_result["quality"] == "hd"
        assert image_result["style"] == "vivid"
        assert image_result["global_size"] == "1024x1024"
        # Should not have text-specific defaults
        assert "max_tokens" not in image_result
        assert "stream_usage" not in image_result
        assert "global_temperature" not in image_result


async def test_transformer_flat_service_defaults_compatibility():
    """Test that flat service default_params work for services that apply to all models."""
    LocalTransformerClient._instance = None

    custom_config = {
        "default_params": {
            "text": {"global_temperature": 0.7},
        },
        "services": {
            "universal_service": {
                "api_key": "${UNIVERSAL_API_KEY}",
                "base_url": "https://api.universal.com/v1",
                "default_params": {
                    "user": "langgate-user",  # Applies to all models
                    "timeout": 30,
                },
            },
        },
        "models": {
            "text": [
                {
                    "id": "universal_service/text-model",
                    "service": {
                        "provider": "universal_service",
                        "model_id": "actual-text-model",
                    },
                }
            ],
        },
        "app_config": {},
    }

    with (
        mock.patch.dict(os.environ, {"UNIVERSAL_API_KEY": "test-key"}),
        patch_load_yaml_config(ConfigSchema.model_validate(custom_config)),
    ):
        client = LocalTransformerClient()

        # Test text model gets flat service defaults
        _, text_result = await client.get_params("universal_service/text-model", {})
        assert text_result["user"] == "langgate-user"
        assert text_result["timeout"] == 30
        assert text_result["global_temperature"] == 0.7


@pytest.mark.asyncio
async def test_transformer_model_specific_overrides(
    local_transformer_client: LocalTransformerClient,
):
    """Test applying model-specific override parameters."""
    # Model with specific thinking override
    _, result = await local_transformer_client.get_params(
        "anthropic/claude-sonnet-4-reasoning", {}
    )

    # Check model-specific override is applied
    assert result["thinking"]["budget_tokens"] == 1024


@pytest.mark.asyncio
async def test_transformer_user_params_precedence(
    local_transformer_client: LocalTransformerClient,
):
    """Test that user parameters have precedence over defaults."""
    # User params should override defaults
    _, result = await local_transformer_client.get_params(
        "gpt-4o", {"temperature": 0.5, "max_tokens": 500}
    )
    assert result["temperature"] == 0.5  # User specified
    assert result["max_tokens"] == 500  # User specified


@pytest.mark.asyncio
async def test_transformer_api_key_resolution(
    local_transformer_client: LocalTransformerClient,
):
    """Test that API keys are resolved from environment variables."""
    # OpenAI API key
    _, result = await local_transformer_client.get_params("gpt-4o", {})
    assert isinstance(result["api_key"], SecretStr)
    assert result["api_key"].get_secret_value() == "sk-test-123"

    # Anthropic API key
    _, result = await local_transformer_client.get_params(
        "anthropic/claude-sonnet-4", {}
    )
    assert isinstance(result["api_key"], SecretStr)
    assert result["api_key"].get_secret_value() == "sk-ant-test-123"


@pytest.mark.asyncio
async def test_transformer_model_specific_renames(
    local_transformer_client: LocalTransformerClient,
):
    """Test applying model-specific parameter renaming."""
    # The claude-sonnet-4 model has stop -> stop_sequences rename
    user_params = {"stop": ["END"], "temperature": 0.7}
    _, result = await local_transformer_client.get_params(
        "anthropic/claude-sonnet-4", user_params
    )

    # Check parameter was renamed
    assert "stop" not in result
    assert "stop_sequences" in result
    assert result["stop_sequences"] == ["END"]


@pytest.mark.asyncio
async def test_transformer_remove_params(
    local_transformer_client: LocalTransformerClient,
):
    """Test removing parameters based on configuration."""
    # The anthropic/claude-sonnet-4 model removes response_format and reasoning
    user_params = {
        "response_format": {"type": "json_object"},
        "reasoning": {"depth": "deep"},
        "presence_penalty": 0.2,
    }
    _, result = await local_transformer_client.get_params(
        "anthropic/claude-sonnet-4", user_params
    )

    # These params should be removed
    assert "response_format" not in result
    assert "reasoning" not in result
    # presence_penalty should remain
    if "presence_penalty" in result:
        assert result["presence_penalty"] == 0.2


@pytest.mark.asyncio
async def test_transformer_invalid_model(
    local_transformer_client: LocalTransformerClient,
):
    """Test error handling for invalid model ID."""
    invalid_model_id = "invalid-model"
    with pytest.raises(
        ValueError, match=f"Model '{invalid_model_id}' not found in configuration"
    ):
        await local_transformer_client.get_params(invalid_model_id, {})


@pytest.mark.asyncio
async def test_transformer_without_env_file():
    """Test that transformer works when .env file doesn't exist."""
    # Reset singleton
    LocalTransformerClient._instance = None

    with (
        mock.patch("pathlib.Path.exists", return_value=False),
        mock.patch.dict(os.environ, {"OPENAI_API_KEY": "direct-env-var"}),
        patch_load_yaml_config(),
    ):
        # Create client with non-existent files
        client = LocalTransformerClient()

        # Should still work with environment variables directly
        model_id = "gpt-4o"
        with pytest.raises(
            ValueError, match=f"Model '{model_id}' not found in configuration"
        ):
            await client.get_params(model_id, {})


# API Format Mapping Tests


@pytest.mark.asyncio
async def test_api_format_service_level_openai_providers(
    local_transformer_client: LocalTransformerClient,
):
    """Test API format resolution for OpenAI-compatible service providers."""
    # Test OpenRouter (uses OpenAI API format)
    api_format, _ = await local_transformer_client.get_params(
        "google/gemma-3-27b-it", {}
    )
    assert api_format == "openai"

    # Test xAI (uses OpenAI API format)
    api_format, _ = await local_transformer_client.get_params("xai/grok-3", {})
    assert api_format == "openai"

    # Test Fireworks AI (uses OpenAI API format)
    api_format, _ = await local_transformer_client.get_params(
        "deepseek/deepseek-r1", {}
    )
    assert api_format == "openai"


@pytest.mark.asyncio
async def test_api_format_fallback_to_provider_name(
    local_transformer_client: LocalTransformerClient,
):
    """Test API format fallback to provider name when no api_format specified."""
    # OpenAI provider (no explicit api_format, should fallback to 'openai')
    api_format, _ = await local_transformer_client.get_params("gpt-4o", {})
    assert api_format == "openai"

    # Anthropic provider (no explicit api_format, should fallback to 'anthropic')
    api_format, _ = await local_transformer_client.get_params(
        "anthropic/claude-sonnet-4", {}
    )
    assert api_format == "anthropic"

    # Gemini provider (no explicit api_format, should fallback to 'gemini')
    api_format, _ = await local_transformer_client.get_params(
        "google/gemini-2.5-pro", {}
    )
    assert api_format == "gemini"

    # MistralAI provider (no explicit api_format, should fallback to 'mistralai')
    api_format, _ = await local_transformer_client.get_params(
        "mistralai/magistral-medium-latest", {}
    )
    assert api_format == "mistralai"


@pytest.mark.asyncio
async def test_api_format_precedence_hierarchy():
    """Test API format precedence: model.api_format → service.api_format → service.provider."""
    # Reset singleton to test with custom configuration
    LocalTransformerClient._instance = None

    # Custom configuration to test precedence
    custom_config = {
        "default_params": {"text": {"temperature": 0.7}},
        "services": {
            "custom_service": {
                "api_key": "${CUSTOM_API_KEY}",
                "base_url": "https://api.custom.com/v1",
                "api_format": "service_format",  # Service-level API format
                "default_params": {},
            }
        },
        "models": {
            "text": [
                {
                    "id": "test/model-with-override",
                    "service": {
                        "provider": "custom_service",
                        "model_id": "actual-model",
                    },
                    "api_format": "model_override",  # Model-level override (highest precedence)
                    "default_params": {},
                    "override_params": {},
                    "remove_params": [],
                    "rename_params": {},
                },
                {
                    "id": "test/model-without-override",
                    "service": {
                        "provider": "custom_service",
                        "model_id": "another-model",
                    },
                    # No model-level api_format, should use service-level
                    "default_params": {},
                    "override_params": {},
                    "remove_params": [],
                    "rename_params": {},
                },
            ]
        },
        "app_config": {},
    }

    with (
        mock.patch.dict(os.environ, {"CUSTOM_API_KEY": "test-key"}),
        patch_load_yaml_config(ConfigSchema.model_validate(custom_config)),
    ):
        client = LocalTransformerClient()

        # Test model-level override (highest precedence)
        api_format, _ = await client.get_params("test/model-with-override", {})
        assert api_format == "model_override"

        # Test service-level fallback (when no model-level override)
        api_format, _ = await client.get_params("test/model-without-override", {})
        assert api_format == "service_format"


@pytest.mark.asyncio
async def test_api_format_provider_name_fallback():
    """Test fallback to provider name when no api_format is specified at any level."""
    # Reset singleton to test with custom configuration
    LocalTransformerClient._instance = None

    # Custom configuration with no api_format specified
    custom_config = {
        "default_params": {"text": {"temperature": 0.7}},
        "services": {
            "fallback_provider": {
                "api_key": "${FALLBACK_API_KEY}",
                "base_url": "https://api.fallback.com/v1",
                # No api_format specified at service level
                "default_params": {},
            }
        },
        "models": {
            "text": [
                {
                    "id": "test/fallback-model",
                    "service": {
                        "provider": "fallback_provider",
                        "model_id": "fallback-model",
                    },
                    # No api_format specified at model level
                    "default_params": {},
                    "override_params": {},
                    "remove_params": [],
                    "rename_params": {},
                },
            ]
        },
        "app_config": {},
    }

    with (
        mock.patch.dict(os.environ, {"FALLBACK_API_KEY": "test-key"}),
        patch_load_yaml_config(ConfigSchema.model_validate(custom_config)),
    ):
        client = LocalTransformerClient()

        # Should fallback to provider name
        api_format, _ = await client.get_params("test/fallback-model", {})
        assert api_format == "fallback_provider"



"""Tests for environment configuration."""

import pytest

from langgate.core.models import ImageModelInfo, LLMInfo
from langgate.registry.models import ModelRegistry
from tests.utils.registry_utils import (
    patch_model_registry,
)


def test_model_registry_singleton_instance():
    """Test that ModelRegistry is a singleton."""
    with patch_model_registry():
        registry1 = ModelRegistry()
        # Create second instance
        registry2 = ModelRegistry()
        # Test singleton behavior
        assert registry1 is registry2
        # Both instances should share the same state
        assert id(registry1._model_caches) == id(registry2._model_caches)


def test_model_registry_list_llms(model_registry: ModelRegistry):
    """Test listing LLMs from the registry."""
    llms = model_registry.list_llms()

    assert isinstance(llms, list)
    assert len(llms) > 0

    for llm in llms:
        assert isinstance(llm, LLMInfo)
        assert llm.id is not None
        assert llm.name is not None
        assert llm.provider is not None
        assert llm.costs is not None


def test_model_registry_get_llm_info(model_registry: ModelRegistry):
    """Test getting specific LLM info from the registry."""
    llms = model_registry.list_llms()
    assert len(llms) > 0

    # Test getting first LLM
    first_llm = llms[0]
    llm_info = model_registry.get_llm_info(first_llm.id)

    assert isinstance(llm_info, LLMInfo)
    assert llm_info.id == first_llm.id
    assert llm_info.name == first_llm.name

    # Test non-existent LLM
    with pytest.raises(ValueError, match="not found"):
        model_registry.get_llm_info("non-existent-llm")


# Image model tests


def test_model_registry_list_image_models(model_registry: ModelRegistry):
    """Test listing image models from the registry."""
    image_models = model_registry.list_image_models()

    assert isinstance(image_models, list)
    assert len(image_models) > 0

    for model in image_models:
        assert isinstance(model, ImageModelInfo)
        assert model.id is not None
        assert model.name is not None
        assert model.provider is not None
        assert model.costs is not None
        assert model.costs.image_generation is not None


def test_model_registry_get_image_model_info(model_registry: ModelRegistry):
    """Test getting specific image model info from the registry."""
    image_models = model_registry.list_image_models()
    assert len(image_models) > 0

    # Test getting first image model
    first_model = image_models[0]
    model_info = model_registry.get_image_model_info(first_model.id)

    assert isinstance(model_info, ImageModelInfo)
    assert model_info.id == first_model.id
    assert model_info.name == first_model.name
    assert model_info.costs.image_generation is not None

    # Test non-existent image model
    with pytest.raises(ValueError, match="not found"):
        model_registry.get_image_model_info("non-existent-image-model")


def test_model_registry_image_model_costs_validation(model_registry: ModelRegistry):
    """Test that image model costs follow the validation rules."""
    image_models = model_registry.list_image_models()
    assert len(image_models) > 0

    for model in image_models:
        costs = model.costs.image_generation

        # Verify exactly one pricing model is set (as per business rule)
        pricing_models = [
            costs.flat_rate,
            costs.quality_tiers,
            costs.cost_per_megapixel,
            costs.cost_per_second,
        ]
        non_null_models = [p for p in pricing_models if p is not None]
        assert len(non_null_models) == 1, (
            f"Model {model.id} should have exactly 1 pricing model, got {len(non_null_models)}"
        )


def test_model_registry_generic_methods(model_registry: ModelRegistry):
    """Test generic get_model_info and list_models methods with modality."""
    # Test text modality (LLMs)
    text_models = model_registry.list_models("text")
    assert isinstance(text_models, list)
    assert len(text_models) > 0
    assert all(isinstance(model, LLMInfo) for model in text_models)

    first_text_model = text_models[0]
    text_model_info = model_registry.get_model_info(first_text_model.id, "text")
    assert isinstance(text_model_info, LLMInfo)
    assert text_model_info.id == first_text_model.id

    # Test image modality
    image_models = model_registry.list_models("image")
    assert isinstance(image_models, list)
    assert len(image_models) > 0
    assert all(isinstance(model, ImageModelInfo) for model in image_models)

    first_image_model = image_models[0]
    image_model_info = model_registry.get_model_info(first_image_model.id, "image")
    assert isinstance(image_model_info, ImageModelInfo)
    assert image_model_info.id == first_image_model.id

    # Test invalid modality
    with pytest.raises(ValueError):
        model_registry.list_models("invalid")  # type: ignore[call-overload]

    with pytest.raises(ValueError):
        model_registry.get_model_info("any-id", "invalid")  # type: ignore[call-overload]


def test_model_registry_modality_separation(model_registry: ModelRegistry):
    """Test that LLMs and image models are properly separated by modality."""
    llms = model_registry.list_llms()
    image_models = model_registry.list_image_models()

    # Get all model IDs
    llm_ids = {llm.id for llm in llms}
    image_model_ids = {model.id for model in image_models}

    # Verify no overlap between LLM and image model IDs
    assert len(llm_ids.intersection(image_model_ids)) == 0, (
        "LLM and image model IDs should not overlap"
    )

    # Verify we can access models through their respective methods
    for llm_id in llm_ids:
        llm = model_registry.get_llm_info(llm_id)
        assert isinstance(llm, LLMInfo)

        # Should not be accessible via image model method
        with pytest.raises(ValueError, match="not found"):
            model_registry.get_image_model_info(llm_id)

    for image_model_id in image_model_ids:
        image_model = model_registry.get_image_model_info(image_model_id)
        assert isinstance(image_model, ImageModelInfo)

        # Should not be accessible via LLM method
        with pytest.raises(ValueError, match="not found"):
            model_registry.get_llm_info(image_model_id)



"""Unit tests for the parameter transformer."""

import os
from unittest import mock

from langgate.transform import ParamTransformer


def test_param_transformer_defaults():
    """Test applying default parameters."""
    transformer = ParamTransformer()
    transformer.with_defaults({"temperature": 0.7, "max_tokens": 1000})

    # Empty user params should get defaults
    result = transformer.transform({})
    assert result["temperature"] == 0.7
    assert result["max_tokens"] == 1000

    # User params should override defaults
    result = transformer.transform({"temperature": 0.5})
    assert result["temperature"] == 0.5
    assert result["max_tokens"] == 1000


def test_param_transformer_overrides():
    """Test applying override parameters."""
    transformer = ParamTransformer()
    transformer.with_defaults({"temperature": 0.7, "max_tokens": 1000})
    transformer.with_overrides({"max_tokens": 2000, "top_p": 0.9})

    # Overrides should take precedence over defaults and user params
    result = transformer.transform({"temperature": 0.5, "max_tokens": 500})
    assert result["temperature"] == 0.5  # User param not in overrides
    assert result["max_tokens"] == 2000  # Override takes precedence
    assert result["top_p"] == 0.9  # Override added


def test_param_transformer_removing():
    """Test removing parameters."""
    transformer = ParamTransformer()
    transformer.with_defaults({"temperature": 0.7, "max_tokens": 1000, "top_p": 0.9})
    transformer.removing(["top_p"])

    # Removed params should not appear in result
    result = transformer.transform({"temperature": 0.5, "top_p": 0.8})
    assert result["temperature"] == 0.5
    assert result["max_tokens"] == 1000
    assert "top_p" not in result


def test_param_transformer_renaming():
    """Test renaming parameters."""
    transformer = ParamTransformer()
    transformer.with_defaults({"temperature": 0.7, "max_tokens": 1000})
    transformer.renaming({"temperature": "temp", "max_tokens": "maxOutputTokens"})

    # Params should be renamed in result
    result = transformer.transform({"temperature": 0.5, "max_tokens": 500})
    assert "temperature" not in result
    assert "max_tokens" not in result
    assert result["temp"] == 0.5
    assert result["maxOutputTokens"] == 500


def test_param_transformer_model_id():
    """Test setting model ID."""
    transformer = ParamTransformer()
    transformer.with_model_id("gpt-4o")

    # Model ID should be set in result
    result = transformer.transform({})
    assert result["model"] == "gpt-4o"

    # User-provided model should be overridden
    result = transformer.transform({"model": "gpt-3.5-turbo"})
    assert result["model"] == "gpt-4o"


def test_param_transformer_env_vars():
    """Test substituting environment variables."""
    with mock.patch.dict(
        os.environ, {"API_KEY": "test-key", "BASE_URL": "http://test.com"}
    ):
        transformer = ParamTransformer()
        transformer.with_defaults(
            {
                "api_key": "${API_KEY}",
                "base_url": "${BASE_URL}",
                "other": "${MISSING_VAR}",
            }
        )
        transformer.with_env_vars()

        result = transformer.transform({})
        assert result["api_key"] == "test-key"
        assert result["base_url"] == "http://test.com"
        assert (
            result["other"] == "${MISSING_VAR}"
        )  # Missing vars should remain as placeholders


def test_param_transformer_chaining():
    """Test chaining multiple transformations."""
    with mock.patch.dict(os.environ, {"API_KEY": "test-key"}):
        transformer = ParamTransformer()

        # Chain multiple transformations
        transformer.with_defaults(
            {
                "temperature": 0.7,
                "max_tokens": 1000,
                "api_key": "${API_KEY}",
            }
        )
        transformer.with_overrides({"max_tokens": 2000})
        transformer.removing(["frequency_penalty"])
        transformer.renaming({"temperature": "temp"})
        transformer.with_model_id("gpt-4o")
        transformer.with_env_vars()

        # Apply all transformations
        result = transformer.transform(
            {
                "temperature": 0.5,
                "frequency_penalty": 0.2,
                "presence_penalty": 0.1,
            }
        )

        # Check result
        assert "temperature" not in result
        assert result["temp"] == 0.5
        assert result["max_tokens"] == 2000
        assert result["model"] == "gpt-4o"
        assert result["api_key"] == "test-key"
        assert "frequency_penalty" not in result
        assert result["presence_penalty"] == 0.1



"""HTTP client for LangGate API."""

from contextlib import asynccontextmanager
from datetime import timedelta
from typing import Generic, get_args

import httpx
from pydantic import SecretStr

from langgate.client.protocol import BaseRegistryClient, ImageInfoT, LLMInfoT
from langgate.core.logging import get_logger
from langgate.core.models import ImageModelInfo, LLMInfo

logger = get_logger(__name__)


def create_registry_http_client(
    base_url: str,
    api_key: SecretStr | None = None,
    timeout: float | httpx.Timeout | None = 10.0,
    **kwargs,
) -> httpx.AsyncClient:
    """
    Creates and configures an httpx.AsyncClient for the Registry API.
    """
    headers = kwargs.pop("headers", {})
    if api_key:
        headers["X-API-Key"] = api_key.get_secret_value()

    processed_base_url = base_url.rstrip("/")

    return httpx.AsyncClient(
        base_url=processed_base_url,
        headers=headers,
        timeout=timeout,
        **kwargs,
    )


class BaseHTTPRegistryClient(
    BaseRegistryClient[LLMInfoT, ImageInfoT], Generic[LLMInfoT, ImageInfoT]
):
    """
    Base HTTP client for the Model Registry API.
    Supports LLMInfo-derived and ImageModelInfo-derived schemas for response parsing and httpx client injection.

    Handles infrequent HTTP requests via temporary clients by default or uses an
    injected client as the request engine. Configuration (base_url, api_key)
    stored in this instance is always used for requests.

    Type Parameters:
        LLMInfoT: The LLMInfo-derived model class for response parsing
        ImageInfoT: The ImageModelInfo-derived model class for response parsing
    """

    __orig_bases__: tuple
    llm_info_cls: type[LLMInfoT]
    image_info_cls: type[ImageInfoT]

    def __init__(
        self,
        base_url: str,
        api_key: SecretStr | None = None,
        cache_ttl: timedelta | None = None,
        llm_info_cls: type[LLMInfoT] | None = None,
        image_info_cls: type[ImageInfoT] | None = None,
        http_client: httpx.AsyncClient | None = None,
    ):
        """Initialize the client.
        Args:
            base_url: The base URL of the registry service
            api_key: Registry server API key for authentication
            cache_ttl: Cache time-to-live
            llm_info_cls: Override for LLM info class
            image_info_cls: Override for image model info class
            http_client: Optional HTTP client for making requests
        """
        super().__init__(cache_ttl=cache_ttl)
        self.base_url = base_url.rstrip("/")
        self.api_key = api_key
        self._http_client = http_client

        # Set model info classes if provided, otherwise they are inferred from the class
        if llm_info_cls is not None:
            self.llm_info_cls = llm_info_cls
        if image_info_cls is not None:
            self.image_info_cls = image_info_cls

        logger.debug(
            "initialized_base_http_registry_client",
            base_url=self.base_url,
            api_key=self.api_key,
            llm_info_cls=getattr(self, "llm_info_cls", None),
            image_info_cls=getattr(self, "image_info_cls", None),
        )

    def __init_subclass__(cls, **kwargs):
        """Set up model classes when this class is subclassed."""
        super().__init_subclass__(**kwargs)

        # Extract the model classes from generic parameters
        if not hasattr(cls, "llm_info_cls") or not hasattr(cls, "image_info_cls"):
            llm_cls, image_cls = cls._get_model_info_classes()
            if not hasattr(cls, "llm_info_cls"):
                cls.llm_info_cls = llm_cls
            if not hasattr(cls, "image_info_cls"):
                cls.image_info_cls = image_cls

    @classmethod
    def _get_model_info_classes(cls) -> tuple[type[LLMInfoT], type[ImageInfoT]]:
        """Extract the model classes from generic type parameters."""
        args = get_args(cls.__orig_bases__[0])
        return args[0], args[1]

    @asynccontextmanager
    async def _get_client_for_request(self):
        """Provides the httpx client to use (injected or temporary)."""
        if self._http_client:
            yield self._http_client
        else:
            async with httpx.AsyncClient() as temp_client:
                yield temp_client

    async def _request(self, method: str, url_path: str, **kwargs) -> httpx.Response:
        """Makes an HTTP request using the appropriate client engine."""
        url = f"{self.base_url}{url_path}"
        headers = kwargs.pop("headers", {})
        if self.api_key:
            headers["X-API-Key"] = self.api_key.get_secret_value()

        async with self._get_client_for_request() as client:
            response = await client.request(method, url, headers=headers, **kwargs)
        return response

    async def _fetch_llm_info(self, model_id: str) -> LLMInfoT:
        """Fetch LLM info from the source via HTTP."""
        response = await self._request("GET", f"/models/llms/{model_id}")
        response.raise_for_status()
        return self.llm_info_cls.model_validate(response.json())

    async def _fetch_image_model_info(self, model_id: str) -> ImageInfoT:
        """Fetch image model info from the source via HTTP."""
        response = await self._request("GET", f"/models/images/{model_id}")
        response.raise_for_status()
        return self.image_info_cls.model_validate(response.json())

    async def _fetch_all_llms(self) -> list[LLMInfoT]:
        """Fetch all LLMs from the source via HTTP."""
        response = await self._request("GET", "/models/llms")
        response.raise_for_status()
        return [self.llm_info_cls.model_validate(model) for model in response.json()]

    async def _fetch_all_image_models(self) -> list[ImageInfoT]:
        """Fetch all image models from the source via HTTP."""
        response = await self._request("GET", "/models/images")
        response.raise_for_status()
        return [self.image_info_cls.model_validate(model) for model in response.json()]


class HTTPRegistryClient(BaseHTTPRegistryClient[LLMInfo, ImageModelInfo]):
    """HTTP client singleton for the Model Registry API using the default schemas."""

    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            logger.debug("creating_http_registry_client_singleton")
        return cls._instance

    def __init__(
        self,
        base_url: str,
        api_key: SecretStr | None = None,
        cache_ttl: timedelta | None = None,
        http_client: httpx.AsyncClient | None = None,
    ):
        if not hasattr(self, "_initialized"):
            super().__init__(base_url, api_key, cache_ttl, http_client=http_client)
            self._initialized = True
            logger.debug("initialized_http_registry_client_singleton")



"""Unit tests for ImageGenerationCost model validation."""

from decimal import Decimal

import pytest

from langgate.core.models import ImageGenerationCost


class TestImageGenerationCostValidation:
    """Test ImageGenerationCost pricing model validation logic."""

    def test_flat_rate_valid(self):
        """Test valid flat rate pricing model."""
        cost = ImageGenerationCost(flat_rate=Decimal("0.05"))
        assert cost.flat_rate == Decimal("0.05")
        assert cost.quality_tiers is None
        assert cost.cost_per_megapixel is None
        assert cost.cost_per_second is None

    def test_quality_tiers_valid(self):
        """Test valid quality tiers pricing model."""
        tiers = {
            "standard": {"input_cost_per_image": Decimal("0.04")},
            "hd": {"input_cost_per_image": Decimal("0.08")},
        }
        cost = ImageGenerationCost(quality_tiers=tiers)
        assert cost.quality_tiers == tiers
        assert cost.flat_rate is None
        assert cost.cost_per_megapixel is None
        assert cost.cost_per_second is None

    def test_cost_per_megapixel_valid(self):
        """Test valid cost per megapixel pricing model."""
        cost = ImageGenerationCost(cost_per_megapixel=Decimal("0.002"))
        assert cost.cost_per_megapixel == Decimal("0.002")
        assert cost.flat_rate is None
        assert cost.quality_tiers is None
        assert cost.cost_per_second is None

    def test_cost_per_second_valid(self):
        """Test valid cost per second pricing model."""
        cost = ImageGenerationCost(cost_per_second=Decimal("0.00025"))
        assert cost.cost_per_second == Decimal("0.00025")
        assert cost.flat_rate is None
        assert cost.quality_tiers is None
        assert cost.cost_per_megapixel is None

    def test_no_pricing_model_raises(self):
        """Test that no pricing model raises ValueError."""
        with pytest.raises(ValueError, match="Exactly one pricing model must be set"):
            ImageGenerationCost()

    def test_multiple_pricing_models_raises(self):
        """Test that multiple pricing models raise ValueError."""
        with pytest.raises(ValueError, match="Exactly one pricing model must be set"):
            ImageGenerationCost(
                flat_rate=Decimal("0.05"), cost_per_second=Decimal("0.01")
            )

    @pytest.mark.parametrize(
        "pricing_data",
        [
            # Two pricing models
            {"flat_rate": Decimal("0.05"), "cost_per_megapixel": Decimal("0.002")},
            # Three pricing models
            {
                "flat_rate": Decimal("0.05"),
                "cost_per_second": Decimal("0.01"),
                "cost_per_megapixel": Decimal("0.002"),
            },
            # All four pricing models
            {
                "flat_rate": Decimal("0.05"),
                "quality_tiers": {"standard": {"cost": Decimal("0.04")}},
                "cost_per_megapixel": Decimal("0.002"),
                "cost_per_second": Decimal("0.01"),
            },
        ],
    )
    def test_multiple_pricing_model_combinations_raise(self, pricing_data):
        """Test various combinations of multiple pricing models all raise ValueError."""
        with pytest.raises(ValueError, match="Exactly one pricing model must be set"):
            ImageGenerationCost(**pricing_data)

    def test_zero_values_valid(self):
        """Test that zero values are valid for pricing models."""
        cost = ImageGenerationCost(flat_rate=Decimal("0"))
        assert cost.flat_rate == Decimal("0")

    def test_empty_quality_tiers_valid(self):
        """Test that empty quality tiers dict is valid."""
        cost = ImageGenerationCost(quality_tiers={})
        assert cost.quality_tiers == {}



"""Local transformer client implementation."""

import importlib.resources
import os
from pathlib import Path
from typing import Any

from dotenv import load_dotenv
from pydantic import SecretStr

from langgate.core.logging import get_logger
from langgate.core.models import Modality
from langgate.core.schemas.config import ConfigSchema, ModelConfig
from langgate.core.utils.config_utils import load_yaml_config, resolve_path
from langgate.transform.protocol import TransformerClientProtocol
from langgate.transform.transformer import ParamTransformer

logger = get_logger(__name__)


class LocalTransformerClient(TransformerClientProtocol):
    """
    Local transformer client for parameter transformations.

    This client handles parameter transformations based on local configuration.
    """

    _instance = None
    _initialized: bool = False

    def __new__(cls, *args, **kwargs):
        """Create or return the singleton instance."""
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(
        self, config_path: Path | None = None, env_file_path: Path | None = None
    ):
        """Initialize the client.

        Args:
            config_path: Path to the configuration file
            env_file_path: Path to the environment file
        """
        if hasattr(self, "_initialized") and self._initialized:
            return

        # Set up default paths
        cwd = Path.cwd()

        core_resources = importlib.resources.files("langgate.core")
        default_config_path = Path(
            str(core_resources.joinpath("data", "default_config.yaml"))
        )
        cwd_config_path = cwd / "langgate_config.yaml"
        cwd_env_path = cwd / ".env"

        self.config_path = resolve_path(
            "LANGGATE_CONFIG",
            config_path,
            cwd_config_path if cwd_config_path.exists() else default_config_path,
            "config_path",
            logger,
        )
        self.env_file_path = resolve_path(
            "LANGGATE_ENV_FILE",
            env_file_path,
            cwd_env_path,
            "env_file_path",
            logger,
        )

        # Cache for configs
        self._global_config: dict[str, Any] = {}
        self._service_config: dict[str, dict[str, Any]] = {}
        self._model_mappings: dict[str, dict[str, Any]] = {}

        # Load configuration
        self._load_config()
        self._initialized = True
        logger.debug("initialized_local_transformer_client")

    def _load_config(self) -> None:
        """Load configuration from file."""
        config = load_yaml_config(self.config_path, ConfigSchema, logger)

        # Extract validated data
        self._global_config = {"default_params": config.default_params}
        self._service_config = {
            k: v.model_dump(exclude_none=True) for k, v in config.services.items()
        }
        self._process_model_mappings(config.models)

        # Load environment variables from .env file if it exists
        if self.env_file_path.exists():
            load_dotenv(self.env_file_path)
            logger.debug("loaded_transformer_env_file", path=str(self.env_file_path))

    def _process_model_mappings(
        self, models_config: dict[str, list[ModelConfig]]
    ) -> None:
        """Process model mappings from validated configuration.

        Args:
            models_config: Dict of modality to list of model configurations
        """
        self._model_mappings = {}

        for modality, model_list in models_config.items():
            for model_config in model_list:
                model_data = model_config.model_dump(exclude_none=True)
                model_id = model_data["id"]
                service = model_data["service"]

                # Store mapping info including modality
                self._model_mappings[model_id] = {
                    "modality": modality,
                    "service_provider": service["provider"],
                    "service_model_id": service["model_id"],
                    "api_format": model_data.get("api_format"),
                    "default_params": model_data["default_params"],
                    "override_params": model_data["override_params"],
                    "remove_params": model_data["remove_params"],
                    "rename_params": model_data["rename_params"],
                }

    async def get_params(
        self, model_id: str, input_params: dict[str, Any]
    ) -> tuple[str, dict[str, Any]]:
        """
        Get transformed parameters for the specified model.

        Parameter precedence and transformation order:

        Defaults (applied only if key doesn't exist yet):
        1. Model-specific defaults (highest precedence for defaults)
        2. Pattern defaults (matching patterns applied in config order)
        3. Service provider defaults
        4. Global defaults (lowest precedence for defaults)

        Overrides/Removals/Renames (applied in order, later steps overwrite/modify earlier ones):
        1. Input parameters (initial state)
        2. Service-level API keys and base URLs
        3. Service-level overrides, removals, renames
        4. Pattern-level overrides, removals, renames (matching patterns applied in config order)
        5. Model-specific overrides, removals, renames (highest precedence)
        6. Model ID (always overwritten with service_model_id)
        7. Environment variable substitution (applied last to all string values)

        Args:
            model_id: The ID of the model to get transformed parameters for
            input_params: The parameters to transform

        Returns:
            A tuple containing (api_format, transformed_parameters)

        Raises:
            ValueError: If the model is not found in the configuration.
        """
        if model_id not in self._model_mappings:
            logger.error("model_not_found", model_id=model_id)
            raise ValueError(f"Model '{model_id}' not found in configuration")

        mapping = self._model_mappings[model_id]
        service_provider = mapping["service_provider"]
        service_model_id = mapping["service_model_id"]
        service_config = self._service_config.get(service_provider, {})

        transformer = ParamTransformer()

        # Collect pattern configurations
        matching_pattern_defaults = []
        matching_pattern_overrides = []
        matching_pattern_removals = []
        matching_pattern_renames = []

        model_patterns = service_config.get("model_patterns", {})
        for pattern, pattern_config in model_patterns.items():
            if pattern in model_id:
                if "default_params" in pattern_config:
                    matching_pattern_defaults.append(pattern_config["default_params"])
                if "override_params" in pattern_config:
                    matching_pattern_overrides.append(pattern_config["override_params"])
                if "remove_params" in pattern_config:
                    matching_pattern_removals.extend(pattern_config["remove_params"])
                if "rename_params" in pattern_config:
                    matching_pattern_renames.append(pattern_config["rename_params"])

        # Apply defaults (highest to lowest precedence)
        transformer.with_defaults(mapping.get("default_params", {}))
        for defaults in reversed(matching_pattern_defaults):
            transformer.with_defaults(defaults)

        # Apply service-level defaults
        service_defaults = service_config.get("default_params", {})
        model_modality = mapping.get("modality")
        if service_defaults:
            # Check if service_defaults is structured by modality
            is_modality_structured = any(
                key in [m.value for m in Modality] for key in service_defaults
            )

            if is_modality_structured and model_modality:
                # Apply modality-specific defaults if available
                modality_defaults = service_defaults.get(model_modality, {})
                if modality_defaults:
                    transformer.with_defaults(modality_defaults)
            elif not is_modality_structured:
                # Apply flat defaults directly
                transformer.with_defaults(service_defaults)

        # Apply modality-specific global defaults
        global_defaults = self._global_config.get("default_params", {})
        modality_defaults = (
            global_defaults.get(model_modality, {}) if model_modality else {}
        )
        transformer.with_defaults(modality_defaults)

        # Apply overrides, removals, renames (service -> pattern -> model precedence)

        # Service level transformations
        api_key_val = service_config.get("api_key")
        if (
            api_key_val
            and isinstance(api_key_val, str)
            and api_key_val.startswith("${")
            and api_key_val.endswith("}")
        ):
            env_var = api_key_val[2:-1]
            env_val = os.environ.get(env_var)
            if env_val is not None:
                transformer.with_overrides({"api_key": SecretStr(env_val)})
            else:
                logger.warning(
                    "api_key_env_var_not_found",
                    variable=env_var,
                    service=service_provider,
                )
        elif api_key_val:
            # Handle non-env var API keys if needed (though config should use env vars)
            await logger.awarning(
                "api_key_set_directly_in_config",
                service=service_provider,
                msg="API key set directly in config, not recommended",
            )
            transformer.with_overrides({"api_key": api_key_val})

        base_url_val = service_config.get("base_url")
        if base_url_val and isinstance(base_url_val, str):
            transformer.with_overrides({"base_url": base_url_val})

        # Service overrides, removals, renames
        service_override_params = service_config.get("override_params", {})
        if isinstance(service_override_params, dict):
            transformer.with_overrides(service_override_params)
        service_remove_params = service_config.get("remove_params", [])
        if isinstance(service_remove_params, list):
            transformer.removing(service_remove_params)
        service_rename_params = service_config.get("rename_params", {})
        if isinstance(service_rename_params, dict):
            transformer.renaming(service_rename_params)

        # Pattern level transformations (apply collected configs)
        for overrides in matching_pattern_overrides:
            transformer.with_overrides(overrides)
        if matching_pattern_removals:
            transformer.removing(matching_pattern_removals)
        for renames in matching_pattern_renames:
            transformer.renaming(renames)

        # Model level transformations (highest precedence for overrides/removals/renames)
        transformer.with_overrides(mapping.get("override_params", {}))
        transformer.removing(mapping.get("remove_params", []))
        transformer.renaming(mapping.get("rename_params", {}))

        # Final transformations
        # Set the final model ID to be sent to the service provider
        transformer.with_model_id(service_model_id)

        # Substitute environment variables in string values
        transformer.with_env_vars()

        # Resolve API format based on fallback hierarchy: model.api_format → service.api_format → service.provider
        api_format = (
            mapping.get("api_format")
            or service_config.get("api_format")
            or service_provider
        )

        # Execute Transformation
        # Applies all chained transformations in order
        transformed_params = transformer.transform(input_params)

        return api_format, transformed_params
